"""
Module EM_MoE.py
================
	Definition of the following ML models useful for developing a full MoE model:
		Mixture of Experts model
			class MoE_model: implements a MoE model with methods for fitting and making predictions
		Softmax regression
			class softmax_regression: implements a softmax regression with methods for fitting and making predictions. This is the standard classifier used by MoE model.
"""
#################

import scipy.stats
import scipy.optimize
import numpy as np
import sys
import os

################# MoE_model class
class MoE_model(object):
	"""
MoE_model
=========
	It represents and fits a MoE model of the form:
		p(y_i | x_i, params) = sum_{k=1}^K S(x_i)_k * N(y_i|<w_k,x_i>+b_k, sigma_k)
	where S(x_i) is any model. A default for S(x_i) is the softmax model S(x_i)= softmax(V*x_i).
	The model is fitted trough EM algorithm.
	"""
	def __init__(self, D, K, gating_function = None, bias = True):
		"""
		Initialize the model.
		The gating function must be specified through an object with methods:
			fit(x_train(N,D)		labels_train (N,K)) returning
			predict((x_test(N,D)))	returning labels_predicted (N,K)
			save					for saving to file the entire model
			load					to load from file the entire model
		Gating function must use cross entropy loss function.
		If None a default softmax regression model is used built from class softmax_regression.
		The gating function obkect must be already initialized properly.
		Input:
			D						dimensionality of input space
			K						number of experts for the model
			gating_function (obj)	an object which represents the gating function
			bias					whether to use a bias in the expert model
		"""
		self.D = D 
		self.K = K 
		if gating_function is None:
			self.gating = softmax_regression(D,K)
		else:
			self.gating = gating_function
		self.W = np.zeros((self.D, self.K)) #[w_1, ..., w_K]
		self.sigma = np.ones((self.K,))
		self.bias = bias
		self.b = np.zeros((self.K,))
		self.initialized = False

	def get_iperparams(self):
		"""
	get_iperparams
	==============
		Returns values of D and K.
		Outpu:
			(D,K)	(dimensionality of input space, number of experts)
		"""
		return (self.D, self.K)

	def save(self, exp_file, gat_file):
		"""
	save
	====
		Saves the model to file.
		Input:
			exp_file	file to save the expert model to
			gat_file	file to save the model for gating function
		"""
		to_save = np.stack((self.b, self.sigma)) #(2, K)
		to_save = np.concatenate((self.W,to_save) , axis = 0) #(D+2,K)
		np.savetxt(exp_file, to_save)
		self.gating.save(gat_file)
		return
	
	def load(self, exp_file, gat_file, load_function = None):
		"""
	load
	====
		Load the model from file. It changes parameters D and K if required.
		Input:
			exp_file		file to load the expert model from
			gat_file		file to load the model for gating function from (using function load_function)
			gating_function	function for loading the gating function model from file
		"""
		weights = np.loadtxt(exp_file)
		self.W = weights[:weights.shape[0]-2,:]
		self.b = weights[weights.shape[0]-2,:]
		self.sigma = weights[weights.shape[0]-1,:]
		if np.all(self.b==0):
			self.bias = False
		else:
			self.bias = True
		self.D = self.W.shape[0]
		self.K = self.W.shape[1]
		
		if load_function is None:
			temp_load = softmax_regression(1,1)
			load_function = temp_load.load

		del self.gating
		self.gating = load_function(gat_file)
		self.initialized =  True
		return

	def experts_predictions(self, X):
		"""
	experts_predictions
	===================
		Returns the predictions of the experts.
		Input:
			X_test (N,D)	test points
		Output:
			y_test (N,K)	experts predictions
		"""
		if X.ndim ==1:
			X = X[:,np.newaxis]
		return np.matmul(X,self.W) + self.b

	def get_gating_probs(self,X):
		"""
	get_gating_probs
	================
		Returns the probability p(z=k|x, params)
		Input:
			X (N,D)	data points
		Output:
			p_gating (N,K)	probabilities of each gating function
		"""
		pi = self.gating.predict(X) #p(z_i = k|x_i) (N,K)
		#pi = np.divide(pi.T, np.sum(pi, axis = 1)).T
		return pi

	def predict(self, X):
		"""
	predict
	=======
		Return the predictions of the model
		Input:
			X (N,D)	test points
		Output:
			y (N,)	model value at test points
		"""
		if X.ndim ==1:
			X = X[:,np.newaxis]
		
		pi = self.gating.predict(X) #p(z_i = k|x_i) (N,K)

		#indices = np.argmax(pi, axis =1)
		#for i in range(pi.shape[0]):
		#	pi[i,indices[i]] = 1.

		pi = np.divide(pi.T, np.sum(pi, axis = 1)).T
		res = np.multiply(pi, self.experts_predictions(X))
		return np.sum(res, axis = 1)

	def expert_likelihood(self, X, y): #give to it a proper name!!!
		"""
	expert_likelihood
	=================
		Computes the quantity p(y_i|x_i, z_i=k) = N(y_i| <w_k,x_i>). This corresponds to the likelihood of each expert for having generated the data.
		Input:
			X (N,D)	data
			y (N,)	targets
		Output:
			pi_k (N,K)	N(y_i| <w_k,x_i>)
		"""
		gaussians_mean = self.experts_predictions(X) #(N,K) X*W + b
		y = np.repeat( np.reshape(y, (len(y),1)), self.K, axis = 1) #(N,K)

		#print('sigma: ', self.sigma)
		res = scipy.stats.norm.pdf( np.divide((y - gaussians_mean), self.sigma) ) #(N,K)
		return np.divide(res, self.sigma) #normalizing result

	def log_likelihood(self, X, y):
		"""
	log_likelihood
	==============
		Computes the log_likelihood for the data given with formula: (Are you sure of the formula??)
			LL 	= sum_{i=1}^N log sum_{k=1}^K p(y_i|x_i, z_i=k) * p(z_i=k |x_i) =
				= sum_{i=1}^N log sum_{k=1}^K N(y_i | <w_k,x_i>) S(x_i)_k
		Input:
			X (N,D)	data
			y (N,)	targets for regression
		Output:
			LL	log-likelihood for the model
		"""
		if X.ndim == 1:
			X = np.reshape(X, (X.shape[0],1))
		exp_likelihood = self.expert_likelihood(X, y)
		res = np.multiply(self.gating.predict(X), exp_likelihood) #(N,K)
		res[np.where(np.abs(res)<1e-30)] = 1e-30 #small regularizer for LL
		res = np.log(np.sum(res,axis=1)) #(N,)
		return np.sum(res) / X.shape[0]

	def __initialise_smart__(self, X, args):
		"""
	__initialise_smart__
	====================
		Having seen the data makes a smart first guess (with farhtest point clustering) for responsibilities and fit gating function model with those responbility.
		Input:
			X (N,D)		train data
			args		arguments to be given to fit method of gating function
		"""
		centroids = np.zeros((self.K,self.D))
		if X.shape[0] > 10*self.K:
			data = X[:10*self.K,:]
		else:
			data = X
		N = data.shape[0]

			#choosing centroids
			#points are chosen from dataset with farhtest point clustering
		ran_index = np.random.choice(N)
		centroids[0,:] = data[ran_index]

		for k in range(1,self.K):
			distances = np.zeros((N,k)) #(N,K)
			for k_prime in range(k):
				distances[:,k_prime] = np.sum(np.square(data - centroids[k_prime,:]), axis =1) #(N,K')
			distances = np.min(distances, axis = 1) #(N,)
			distances /= np.sum(distances) #normalizing distances to make it a prob vector
			next_cl_arg = np.random.choice(range(data.shape[0]), p = distances) #chosen argument for the next cluster center
			centroids[k,:] = data[next_cl_arg,:]

		var = np.var(X, axis = 0) #(D,)

			#computing initial responsibilities
		r_0 = np.zeros((X.shape[0],self.K))
		for k in range(self.K):
			r_0[:,k] = np.sum(np.divide(np.square(X - centroids[k,:]), var), axis = 1) + 1e-5
		r_0 = np.divide(r_0.T, np.sum(r_0,axis=1)).T

		self.gating.fit(X,r_0, *args)

		return r_0



	def __initialise__(self, X, args):
		"""
	__initialise__
	==============
		Having seen the data makes a first guess for responsibilities and fit gating function model with those responbility.
		Input:
			X (N,D)		train data
			args		arguments to be given to fit method of gating function
		"""
			#getting centroids
		indices = np.random.choice(range(X.shape[0]), size = self.K, replace = False)
		centroids = X[indices,:] #(K,D) #K centroids are chosen
			#getting variances
		var = np.var(X, axis = 0) #(D,)

			#computing initial responsibilities
		r_0 = np.zeros((X.shape[0],self.K))
		for k in range(self.K):
			r_0[:,k] = np.divide(np.square(X - centroids[k,:]), var)[:,0] + 1e-10
		r_0 = np.divide(r_0.T, np.sum(r_0,axis=1)).T

		self.gating.fit(X,r_0, *args)

		return r_0

	def fit(self, X_train, y_train, N_iter=None, threshold = 1e-2, args= [], verbose = False, val_set = None, pick_best = False):
		"""
	fit
	===
		Fit the model using EM algorithm.
		Input:
			X_train (N,D)	train data
			y_train (N,)	train targets for regression
			N_iter			Maximum number of iteration (if None only threshold is applied)
			threshold		Minimum change in LL below which algorithm is terminated
			args			arguments to be given to fit method of gating function
			verbose 		whether to print values during fit
			val_set			tuple (X_val, y_val) with a validation set to test performances
			pick_best		if True the model with best validation mse is chosen as best model (doesn't apply if val_set is None)
		Output:
			history		list of value for the LL of the model at every epoch
		"""
		if X_train.ndim == 1:
			X_train = np.reshape(X_train, (X_train.shape[0],1))
		if X_train.shape[1] != self.D:
			raise TypeError("Wrong shape for X_train matrix "+str(X_train.shape)+". Second dimension should have lenght "+str(self.D))
		if y_train.ndim > 1:
			y_train = y_train[:,0]

		if threshold is None:
			threshold = 0

			#initialization
		if not self.initialized:
			r_0 = self.__initialise_smart__(X_train,args)
			self.EM_step(X_train,y_train, r_0, args)

		i = 0
		old_LL = (-1e5,)
		LL = (-1e4,)
		if pick_best: best_mse = 1e6 #best mse so far (only if val_set is not None)
		history=[]
		while(LL[0] - old_LL[0] > threshold ): #exit condition is decided by train LL (even if val_set is not None): don't know why...
				#do batch update!!!
			old_LL = LL
			gat_history = self.EM_step(X_train,y_train, args = args)
			LL=(self.log_likelihood(X_train, y_train),)
			if isinstance(val_set,tuple) :
				LL = (LL[0], self.log_likelihood(val_set[0], val_set[1]))
			history.append(LL)
			i += 1
			if verbose:
				print("LL at iter "+str(i)+"= ",LL)
				try:
					print("   Gating loss: ", gat_history[0], gat_history[-1])
				except TypeError:
					pass
				if isinstance(val_set,tuple) :
					mse = np.sum(np.square( self.predict(val_set[0])-val_set[1]))/val_set[0].shape[0]
					print("   Val loss: ", mse)
					if pick_best:
						if mse < best_mse:
							best_mse = mse
							print("Chosen the best!")
							try: #saving best model so far
								self.save("temp_exp", "temp_gat") 
							except:
								os.system("rm -f temp_exp temp_gat")
			if N_iter is not None:
				if i>= N_iter:
					break
			try:
				assert LL[0] - old_LL[0] >=0 #train LL must always increase in EM algorithm. Useful check
			except:
				break #if LL increase (and it shouldn't) EM should terminate

		self.initialized =  True

		if isinstance(val_set,tuple) and pick_best: #loading best model so far
			files = os.listdir(".")
			if "temp_exp" in files and  "temp_gat" in files:
				print("loaded the best")
				self.load = ("temp_exp", "temp_gat")
				os.system("rm -f temp_exp temp_gat")

		return history

	def EM_step(self, X, y, r = None, args = []):
		"""
	EM_step
	=======
		Does one EM update.
		Input:
			X (N,D)	train data
			y (N,)	train targets for regression
			r (N,K)	responsibilities for the E step (if None they are computed with method get_responsibilities)
			args	some arguments to pass to fit method of gating function
		Output:
			gat_history		history for the gating function fit
		"""
			#E step
		if r is None:
			r= self.get_responsibilities(X,y) #(N,K)

			#M step
			#M step for experts
				#weights is updated by solving a linear fit with weights r_{ik} in loss function
		if self.bias:
			X_temp = np.concatenate((np.ones((X.shape[0],1)),X), axis = 1)
		else:
			X_temp = X
		for k in range(self.K):
			R = np.diag(r[:,k]) #(N,N)
			temp = np.linalg.inv(np.matmul(X_temp.T,np.matmul(R,X_temp))) #(D,D)/(D+1,D+1)
			temp = np.matmul(np.matmul(temp, X_temp.T),R) #(D,N)/(D+1,N)
			temp = np.matmul(temp, y) #(D,)/(D+1,)
			if self.bias:
				self.b[k] = temp[0] #()
				self.W[:,k] = temp[1:] #(D,)
			else:
				self.W[:,k] = temp #(D,)
			sigma_square = np.sum(np.multiply(r[:,k], np.square(y-self.experts_predictions(X)[:,k])) ) / np.sum(r[:,k])
			self.sigma[k] = np.sqrt(sigma_square)

			#M step for gating functions
		gat_history = self.gating.fit(X,r, *args)
		return gat_history

	def get_responsibilities(self,X,y):
		"""
	get_responsibilities
	====================
		Computes responsibilities for the given input data:
			r_k = p(y=k|x)
		Input:
			X (N,D)	data
			y (N,)	data labels
		"""
		pi = self.gating.predict(X) #p(z_i = k|x_i) (N,K)
		pi = np.divide(pi.T, np.sum(pi, axis = 1)).T
		exp_term = self.expert_likelihood(X, y)

		r =  np.multiply(pi,  exp_term) + 1e-10 # (N,K) responsibilities matrix
		r = np.divide(r.T, np.sum(r, axis =1)).T
		assert np.all(r>=0) 								#checking that all r_{ik} are more than zero
		assert np.all(np.abs(np.sum(r,axis=1) - 1)<1e-5) 	#checking proper normalization

		#r_new = np.zeros(r.shape)
		#print("r",r)
		#indices = np.argmax(r, axis =1)
		#print(indices)
		#for i in range(r.shape[0]):
		#	r_new[i,indices[i]] = 1.
		#print("r_new", r_new)
		return r

################# softmax_regression class
class softmax_regression(object):
	"""
softmax_regression
==================
	Implements a class for softmax regression with K labels. It has the form:
		p(y= k|x,V) = exp( V_k*x ) / sum_{k=1}^K exp( V_k*x + b_k )
	It has methods for getting predictions from the model and to fit the model.
	"""
	def __init__(self, D, K):
		"""
	__init__
	========
		Initialize the model with K classes for regressions.
		Input:
			D 	dimensionality of input space
			K	number or classes for regression
		"""
		self.D = D 
		self.K = K 
		self.V = np.zeros((D+1,K))
		return

	def save(self, filename):
		"""
	save
	====
		Save softmax model to file.
		Input:
			filename	name of the file to save the model to
		"""
		np.savetxt(filename, self.V)
		return

	def load(self, filename):
		"""
	load
	====
		Load the model from file.
		Input:
			filename	name of the file to load the model from
		"""
		self.V = np.loadtxt(filename)
		self.D = self.V.shape[0]-1
		self.K = self.V.shape[1]
		return self

	def predict(self, X_test, V = None):
		"""
	predict
	=======
		Makes predictions for the softmax regression model. Weights can be freely specified by the user.
		Input:
			X_test (N,D)	test points
			V (D,K)			weight of the model that gives predictions (if None, internal weights are used)
		Output:
			y_test (N,K)	model prediction
		"""
		if X_test.ndim == 1:
			X_test = X_test[:,np.newaxis]
		if X_test.shape[1] == self.D:
			X_test = np.concatenate((np.ones((X_test.shape[0],1)), X_test), axis = 1) #adding dummy variable for the bias
		if V is None:
			V = self.V
		res = np.matmul(X_test,V) #(N,K)
		res = np.exp(res) + 1e-5 #(N,K)
		res = np.divide(res.T, np.sum(res, axis = 1)).T #(N,K) normalizing

		return res

	def fit_single_loop(self, X_train, y_train):
		"""
	fit_single_loop
	===============
		######## DOES NOT WORK YET!! WE ARE SORRY FOR THE INCONVENIENT ######
		Fit the model using the closed form of LL of the problem.
		See: https://link.springer.com/content/pdf/10.1007%2F978-3-642-01510-6_109.pdf 
		Input:
			X_train (N,D)	train data
			y_train (N,)	train targets for regression
		"""
		if X_train.shape[1] == self.D:
			X_train = np.concatenate((np.ones((X_train.shape[0],1)), X_train), axis = 1) #adding dummy variable for the bias

		for k in range(self.K-1):
			#print(np.where(y_train[:,-1]+2e-5 ==0))
			div = np.divide(y_train[:,k],y_train[:,-1]+2e-10)
			div[np.where(div ==0)] = 1e-20
			H = np.log(div) # (N,) "new targets for linear regression"
			fitted_V = np.matmul(np.linalg.inv(np.matmul(X_train.T,X_train)), X_train.T) #(D+1,N)
			fitted_V = np.matmul(fitted_V, H)
			self.V[:,k] = fitted_V

		self.V[:,-1] = np.zeros((self.D+1,)) #last is zero by default! Remember it!

		non_zero = np.where(self.predict(X_train)<0)
		print("<0: ",non_zero)
		return 

	def accuracy(self, X_test, y_test):
		"""
	accuracy
	========
		Computes the accuracy of the model (i.e. the fraction of misclassified points).
		This measure is meaningful only in the case of hard clustering where there is only one label for each data point.
		Input:
			X_test (N,D)	test points
			y_test (N,K)	true labels of test points
		Output:
			accuracy	accuracy of the predictions made at test points
		"""
		if X_test.ndim == 1:
			X_test = np.reshape(X_test, (X_test.shape[0],1))
		y_pred = self.predict(X_test)
		return np.sum(np.argmax(y_pred,axis=1)==np.argmax(y_test,axis=1))/float(y_test.shape[0])


	def LL(self, X_test, y_test):
		"""
	LL
	==
		Evaluate the log-likelihood for the model given X and their labels.
		Input:
			X_test (N,D)	test points
			y_test (N,K)	labels of test points
		Output:
			LL	log-likelihood for the model
		"""
		if X_test.ndim == 1:
			X_test = np.reshape(X_test, (X_test.shape[0],1))
		return self.loss(self.V, [X_test, y_test, 0.]) * X_test.shape[0]
		

	def loss(self, V, data):
		"""
	loss
	====
		Loss function to minimize wrt. V. It is the function:
			NLL(V) = - log[p(D|V)]
		Input:
			V ((D+1)*K,)	weights of logreg
			data 			list [X_train (N,D), y_train (N,K), lambda ()]
		Output:
			loss	value for the loss function evaluated at V	
		"""
		V = np.reshape(V, (self.D+1,self.K))
		X = data[0]
		y = data[1]
		reg_constant = data[2]

		mu = self.predict(X, V) #(N,K)
			#mu must be regularized in logarithm. Otherwise it might give Nan if a label prob is 0
		LL = -(np.sum(np.multiply(y, np.log(mu+1e-40))) / X.shape[0]) + 0.5* reg_constant * np.sum(np.square(V))
		return LL

	def grad(self, V, data):
		"""
	grad
	====
		Gradient of the loss function to minimize wrt. V (see function loss())
		Input:
			V ((D+1)*K,)	weights of logreg
			data 			list [X_train (N,D), y_train (N,K), lambda ()]
		Output:
			grad	value for the gradient of loss function evaluated at V	
		"""
		to_reshape = False
		if V.ndim == 1:
			V = np.reshape(V, (self.D+1,self.K))
			to_reshape = True
		X = data[0]
		y = data[1]
		reg_constant = data[2] #regularizer
		
		mu = self.predict(X, V) #(N,K)
		delta = (mu - y) + 1e-4
		grad = np.matmul(X.T,delta) / X.shape[0] + reg_constant*V #(N,D).T (N,K) = (D,K)
		if to_reshape:
			return np.reshape(grad, ((self.D+1)*self.K,))
		else:
			return grad
	
	def get_weights(self):
		"""
	get_weights
	===========
		Returns the weights of the model.
		Input:
		Output:
			V (D+1,K)	weights for the model
		"""
		return self.V

	def fit(self, X_train, y_train, opt = "adam", val_set = None, reg_constant = 1e-4, verbose = False, threshold = 1e-2, N_iter = 30, learning_rate = 1e-3):
		"""
	fit
	===
		Fit the model using gradient descent.
		Can use adam for adaptive step: https://arxiv.org/abs/1412.6980v8
		Can use bfgs method provided by scipy: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fmin_bfgs.html
		Using a small regularizer can help convergence (especially for bfgs).
		Input:
			X_train (N,D)	train data
			y_train (N,)	train targets for regression
			opt				which optimizer to use ("adam" or "bsfg")
			val_set			tuple (X_val, y_val) with a validation set to test performances
			reg_constant	regularization constants
			verbose			whether to print values of loss function at every train step
			threshold		minimun improvement of validation erorr on 10 iteration before stopping (train error if val_set =None)
			N_iter			number of iteration to be performed (doesn't apply to bfgs or if threshold is not None)
			learning_rate	learning rate used for gradient update (doesn't apply to bfgs)
		Output:
			history		list of value for the loss function
		"""
		if X_train.ndim == 1:
			X_train = np.reshape(X_train, (X_train.shape[0],1))
		if X_train.shape[1] == self.D:
			X_train = np.concatenate((np.ones((X_train.shape[0],1)), X_train), axis = 1) #not necessary but might be useful to speed up the code

		args = (X_train, y_train, reg_constant) #arguments for loss and gradients
		
		if opt == "adam":
			return self.__optimize_adam__(args, threshold, N_iter, learning_rate, verbose, val_set)
		if opt == "bfgs":
			return self.__optimise_bfgs__(args, verbose, val_set)

	def __optimise_bfgs__(self, args, verbose, val_set = None):
		"""
	__optimise_bfgs__
	=================
		Wrapper to scipy.optimize.fmin_bfgs (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fmin_bfgs.html) to minimise loss function.
		Input:
			args		tuple of arguments to be passed to loss and gradient [X_train (N,D), y_train (N,K), lambda ()]
			verbose		whether to print scipy convergence message
			val_set		tuple (X_val, y_val) with a validation set to test performances
		Output:
			history		initial and final value of loss function
		"""
		loss_0 = (self.loss(self.V, args),)
		if isinstance(val_set,tuple):
			args_val = (val_set[0], val_set[1], args[2])
			loss_val = self.loss(self.V, args_val)
			loss_0 = (loss_0, loss_val)

			#wrapper to self.loss and self.grad to make them suitable for scipy
		loss = lambda V, a,b,c: self.loss(V,(a,b,c))
		grad = lambda V, a,b,c: self.grad(V,(a,b,c))

		res = scipy.optimize.fmin_bfgs(loss, self.V.reshape(((self.D+1)*self.K,)), grad, args , disp = verbose)
		self.V = res.reshape((self.D+1,self.K))

		if isinstance(val_set,tuple):
			loss_fin = (self.loss(self.V, args), self.loss(self.V, args_val))
		else:
			loss_fin = (self.loss(self.V, args), )
		return [loss_0, loss_fin]


	def __optimize_adam__(self, args, threshold, N_iter, learning_rate, verbose, val_set = None):
		"""
	__optimise_adam__
	=================
		Implements optimizer with to perform adaptive step gradient descent.
		The implementation follows: https://arxiv.org/abs/1412.6980v8
		Input:
			args			tuple of arguments to be passed to loss and gradient [X_train (N,D), y_train (N,K), lambda ()]
			threshold		minimun improvement of train erorr before stopping fitting procedure
			N_iter			number of iteration to be performed
			learning_rate	learning rate to be set for adam
			verbose			whether to print loss at each step
			val_set			tuple (X_val, y_val) with a validation set to test performances
		Output:
			history		list of loss function value (train, )/(train,val) at each iteration step
		"""
			#setting parameters for learning rate
		beta1 = .9		#forgetting factor for first moment
		beta2 = .999	#forgetting factor for second moment
		epsilon = 1e-8
		m = np.zeros(self.V.shape) #first moment
		v = np.zeros(self.V.shape) #second moment
		history = []
		if threshold is not None:
			N_iter = 1000000000 # if threshold, no maximum iteration should be used
		for i in range(0,N_iter):
			g = self.grad(self.V, args)
			m = beta1*m + (1-beta1)*g
			v = beta2*v + (1-beta2)*np.square(g)
			m_corr = m / (1-beta1)
			v_corr = v / (1-beta2)

			update = np.divide(m_corr, np.sqrt(v_corr)+epsilon)
			if np.any(np.isnan(update)): #debug
				quit()
			self.V = self.V - learning_rate * update
			self.V[:,-1] = np.zeros((self.D+1,))

			if isinstance(val_set,tuple):
				args_val = (val_set[0], val_set[1], args[2])
				history.append((self.loss(self.V, args), self.loss(self.V, args_val)) ) #(train_err, val_err)
			else:
				history.append((self.loss(self.V, args),))

			if verbose:
				print("Loss at iter= ",i, history[i])

			if threshold is not None and i>10:
				if history[-10][-1] - history[-1][-1] < threshold:
					break
		return history







"""
Module fit_model.py
===================
	Various routines for mlgw model fitting.
	Routines:
		create_PCA_dataset		routine for generating a dataset for PC projections of waves and orbital parameters. It start from a waveform dataset
		fit_MoE					routine for fitting a MoE model for regression from obrital parameters to PC projection. It takes as input a PCA dataset, and outputs many file where the model is saved.
"""

#################

import os
import sys
import warnings
import numpy as np
sys.path.insert(1, os.path.dirname(__file__)) 	#adding to path folder where mlgw package is installed (ugly?)
from shutil import copyfile #for copying files 
from GW_helper import * 	#routines for dealing with datasets
from ML_routines import *	#PCA model
from EM_MoE import *		#MoE model

################# routine create_PCA_dataset
def create_PCA_dataset(K, dataset_file, out_folder, train_frac = 0.75):
	"""
create_PCA_dataset
==================
	Creates a PCA dataset starting from a waveform dataset.
	Waveform dataset can be generated with:
		mlgw.GW_helper.create_dataset_TD(N_data = 3000, N_grid = 4000, filename = "../GW_TD_dataset.dat",
		 t_coal = .4, q_range = (1.,10.), m2_range = None, s1_range = (-0.8,0.8), s2_range = (-0.8,0.8),
		 t_step = 1e-6, lal_approximant = "SEOBNRv2_opt")
	A PCA dataset consists in
		PCA model				a fitted PCA model in the form of mlgw.ML_routines.PCA_model (used for reduced dataset creation)
		train(test) theta		training(test) set for the orbital parameters
		train(test) amp(ph)		training(test) set for the reduced components of amplitude(phase)
		times					times at which waves are evalueted in the high dimensional representation (not useful for MoE but required by mlgw.GW_generator
	Dataset will be saved to output folder in the following files (total 9):
		"amp(ph)_PCA_model"    "PCA_train(test)_theta.dat"    "PCA_train(test)_amp(ph).dat"    "times"
	While loading, amplitude are scaled by a factor of 1e-21 to make them O(1).
	Input:
		K (tuple)		number of PC to consider (K_amp, K_ph); if int amp and ph have the same number of PC
		dataset_file	path to file holding input waveform dataset
		out_folder		output folder which all output files will be saved to.
		train_frac		fraction of data in WF datset to be included in training set (must be strictly less than 1)
	"""
	theta_vector, amp_dataset, ph_dataset, times = load_dataset(dataset_file, shuffle=True) #loading dataset
	print("Loaded datataset with shape: "+ str(ph_dataset.shape))

	train_theta, test_theta, train_amp, test_amp = make_set_split(theta_vector, amp_dataset, train_frac, 1e-21)
	train_theta, test_theta, train_ph, test_ph   = make_set_split(theta_vector, ph_dataset, train_frac, 1.)

	print("Orbital parameters are in range: [%f,%f]x[%f,%f]x[%f,%f]"%(np.min(train_theta[:,0]), np.max(train_theta[:,0]), np.min(train_theta[:,1]), np.max(train_theta[:,1]), np.min(train_theta[:,2]), np.max(train_theta[:,2])))

	if type(K) is int:
		K = (K,K)
	if type(K) is not tuple:
		raise RuntimeError("Wrong format for number of component K. Tuple expected but got "+str(type(K)))

		#DOING PCA

		#amplitude
	PCA_amp = PCA_model()
	E_amp = PCA_amp.fit_model(train_amp, K[0], scale_PC=True)
	print("PCA eigenvalues for amplitude: ", E_amp)
	red_train_amp = PCA_amp.reduce_data(train_amp)			#(N,K) to save in train dataset 
	red_test_amp = PCA_amp.reduce_data(test_amp)			#(N,K) to save in test dataset
	rec_test_amp = PCA_amp.reconstruct_data(red_test_amp) 	#(N,D) for computing mismatch

		#phase
	PCA_ph = PCA_model()
	E_ph = PCA_ph.fit_model(train_ph, K[1], scale_PC=True)
	print("PCA eigenvalues for phase: ", E_ph)
	red_train_ph = PCA_ph.reduce_data(train_ph)			#(N,K) to save in train dataset 
	red_test_ph = PCA_ph.reduce_data(test_ph)			#(N,K) to save in test dataset
	rec_test_ph = PCA_ph.reconstruct_data(red_test_ph) 	#(N,D) for computing mismatch

	if not out_folder.endswith('/'):
		out_folder = out_folder + "/"

		#saving to files
	PCA_amp.save_model(out_folder+"amp_PCA_model")				#saving amp PCA model
	PCA_ph.save_model(out_folder+"ph_PCA_model")				#saving ph PCA model
	np.savetxt(out_folder+"PCA_train_theta.dat", train_theta)	#saving train theta
	np.savetxt(out_folder+"PCA_test_theta.dat", test_theta)		#saving test theta
	np.savetxt(out_folder+"PCA_train_amp.dat", red_train_amp)	#saving train reduced amplitudes
	np.savetxt(out_folder+"PCA_test_amp.dat", red_test_amp)		#saving test reduced amplitudes
	np.savetxt(out_folder+"PCA_train_ph.dat", red_train_ph)		#saving train reduced phases
	np.savetxt(out_folder+"PCA_test_ph.dat", red_test_ph)		#saving test reduced phases
	np.savetxt(out_folder+"times", times)						#saving times

	F_PCA = compute_mismatch(test_amp, test_ph, rec_test_amp, rec_test_ph)
	print("Average PCA mismatch: ",np.mean(F_PCA))
	
	return

################# routine fit_MoE
def fit_MoE(fit_type, in_folder, out_folder, experts, comp_to_fit = None, features = None, EM_threshold = 1e-2, args = None, N_train = None, verbose = True, train_mismatch = False, test_mismatch = True):
	"""
fit_MoE
=======
	Fit a MoE model for each component of a PCA dataset.
	It loads a PCA dataset from in_folder and fits the regression
		theta = (q,s1,s2) ---> PC_pojection(theta)
	Outputs the fitted models to out_folder
		amp(ph)_exp_#		for amplitude (phase) of expert model for PCA component #
		amp(ph)_gat_#		for amplitude (phase) of gating function for PCA component #
		amp(ph)_feat		for list of features to use for MoE models
	Furthermore it copies PCA models and times files to the out folder. Out folder will be a valid input for mlgw.GW_generator.GW_generator.
	User can choose some fitting hyperparameters.
	Input:
		fit_type ("amp","ph")	whether to fit the model for amplitude or phase
		in_folder				path to folder with the PCA dataset. It must have the format of mlgw.fit_model.create_PCA_dataset
		out_folder				path to folder to save models to. The folder can be used by mlgw.GW_generator.GW_generator
		experts ()/[]			experts to use for each PCA component fit. If int, all models have same number of experts; if list, different experts are allowed but they must match the actual number of PC in PCA dataset.
		comp_to_fit	[]			list of PCs to fit. If None, all components will be fitted. If int, it denotes the maximum PC order to be fitted.
		features []				list of feature for basis function expansion. It must be in the format of mlgw.ML_routines.add_extra_features. If None, a default second degree polynomial in q, s1, s2 will be used.
		EM_threshold ()			threshold of minumum change in LL before breaking from the EM algorithm.
		args []					list of arguments for the softmax function fit routine mlgw.EM_MoE.softmax_regression.fit. They must be in the order [optimizator, validation set, reg. constant, verbose, threshold, # iteration, step for gradient]. If None, default values are used (recommended)
		N_train					number of training points to use in the PCA dataset. If None, every point available will be used.
		verbose					whether to display EM iteration messages
		train_mismatch			whether to return mismatch and mse on train data (if True, test_mismatch = True)
		test_mismatch			whether to return mismatch and mse on test data
	Output:
		F, mse_list											average test mismatch with PCA reconstructed waves, list of mse for each of the fitted component
		F_train, F_test, mse_train_list, mse_test_list		average train (test) mismatch (if relevant). Same format as above.
	"""
	if train_mismatch:
		test_mismatch = True

	if not fit_type in ["amp","ph"]:
		raise RuntimeError("Data type for fit_type not understood. Required (\"amp\"/\"ph\") but "+str(fit_type)+" given.")
		return

	if not os.path.isdir(out_folder): #check if out_folder exists
		raise RuntimeError("Ouput folder "+str(out_folder)+" does not exist. Please, choose a valid folder.")
		return

	if not out_folder.endswith('/'):
		out_folder = out_folder + "/"
	if not in_folder.endswith('/'):
		in_folder = in_folder + "/"

	if features is None:
		features = ["00", "11","22", "01", "02", "12"]
	if type(features) is not list:
		raise RuntimeError("Features to use for regression must be given as list. Type "+str(type(features))+" given instead")
		return
	
	if type(N_train) is not int and N_train is not None:
		raise RuntimeError("Nunmber of training point to use must be be an integer. Type "+str(type(N_train))+" given instead")
		return

	if args is None:
				#opt	val_set reg   verbose threshold	N_it step
		args = ["adam", None,   1e-5, False,  1e-4,		150, 2e-3] #default arguments for sotmax fit routine

		#loading data
	train_theta = np.loadtxt(in_folder+"PCA_train_theta.dat")[:N_train,:]		#(N,3)
	test_theta = np.loadtxt(in_folder+"PCA_test_theta.dat")						#(N',3)
	PCA_train = np.loadtxt(in_folder+"PCA_train_"+fit_type+".dat")[:N_train,:]	#(N,K)
	PCA_test = np.loadtxt(in_folder+"PCA_test_"+fit_type+".dat")				#(N',K)
	PCA = PCA_model(in_folder+fit_type+"_PCA_model")							#loading PCA model

	print("Using "+str(PCA_train.shape[0])+" train data")
	
		#adding new features for basis function expansion
	train_theta = add_extra_features(train_theta, features)
	test_theta = add_extra_features(test_theta, features)
	D = train_theta.shape[1] #dimensionality of input space for MoE

	MoE_models = [] #list of model, one for each component
	if train_mismatch:
		PCA_train_pred = np.zeros(PCA_train.shape) #to keep values for reconstruction
	PCA_test_pred = np.zeros(PCA_test.shape) #to keep values for reconstruction

	if comp_to_fit is None:
		comp_to_fit = [i for i in range(PCA.get_dimensions()[1])]
	if type(comp_to_fit) is int:
		comp_to_fit = [i for i in range(comp_to_fit)]
	if type(comp_to_fit) is not list:
		raise RuntimeError("Components to fit must be given as list or None type. Type "+str(type(comp_to_fit))+" given instead")
		return

	if type(experts) is int:
		experts = [experts for i in comp_to_fit]

	mse_train_list = [] #list for holding mse of every PCs
	mse_test_list = [] #list for holding mse of every PCs

		#starting fit procedure
	for k in comp_to_fit:
		print("### Fitting component ", k, " | experts = ", experts[k])
			#useless variables for sake of clariness
		y_train = PCA_train[:,k]
		y_test = PCA_test[:,k]

		MoE_models.append(MoE_model(D,experts[k]))
		MoE_models[-1].fit(train_theta, y_train, threshold = EM_threshold, args = args, verbose = verbose, val_set = (test_theta, y_test))

			#doing some test
		if train_mismatch:
			y_pred = MoE_models[-1].predict(train_theta)
			mse_train_list.append( np.sum(np.square(y_pred-y_train))/(y_pred.shape[0]) )
			PCA_train_pred[:,k] = y_pred

		y_pred = MoE_models[-1].predict(test_theta)
		mse_test_list.append( np.sum(np.square(y_pred-y_test))/(y_pred.shape[0]) )
		print("Test square loss for comp "+str(k)+": ", mse_test_list[-1] )
		print("LL for comp "+str(k)+" (train,val): ", (MoE_models[-1].log_likelihood(train_theta,y_train),MoE_models[-1].log_likelihood(test_theta,y_test)))
		PCA_test_pred[:,k] = y_pred


		#saving everything to file
		#saving feature list
	outfile = open(out_folder+fit_type+"_feat", "w+")
	outfile.write("\n".join(features))
	outfile.close()
		#copying PCA model and times, for making out_folder ready to be used in 
	copyfile(in_folder+fit_type+"_PCA_model", out_folder+fit_type+"_PCA_model")
	copyfile(in_folder+"times", out_folder+"times")
		#saving MoE models
	for i in range(len(MoE_models)):
		MoE_models[i].save(out_folder+fit_type+"_exp_"+str(i),out_folder+fit_type+"_gat_"+str(i))

		#doing test
	if test_mismatch and fit_type is "ph": #testing for phase
		PCA_test_amp = np.loadtxt(in_folder+"PCA_test_amp.dat")
		PCA_amp = PCA_model(in_folder+"amp_PCA_model")
		rec_amp=PCA_amp.reconstruct_data(PCA_test_amp)
		rec_ph=PCA.reconstruct_data(PCA_test)
		rec_ph_pred=PCA.reconstruct_data(PCA_test_pred)
		F_MoE = compute_mismatch(rec_amp, rec_ph, rec_amp, rec_ph_pred)
		
	if test_mismatch and fit_type is "amp": #testing for amplitude
		PCA_test_ph = np.loadtxt(in_folder+"PCA_test_ph.dat")
		PCA_ph = PCA_model(in_folder+"ph_PCA_model")
		rec_ph=PCA_ph.reconstruct_data(PCA_test_ph)
		rec_amp=PCA.reconstruct_data(PCA_test)
		rec_amp_pred=PCA.reconstruct_data(PCA_test_pred)
		F_MoE = compute_mismatch(rec_amp, rec_ph, rec_amp_pred, rec_ph)

	if train_mismatch and fit_type is "ph": #testing for phase
		PCA_train_amp = np.loadtxt(in_folder+"PCA_train_amp.dat")[:N_train,:]
		PCA_amp = PCA_model(in_folder+"amp_PCA_model")
		rec_amp=PCA_amp.reconstruct_data(PCA_train_amp)
		rec_ph=PCA.reconstruct_data(PCA_train)
		rec_ph_pred=PCA.reconstruct_data(PCA_train_pred)
		F_MoE_train = compute_mismatch(rec_amp, rec_ph, rec_amp, rec_ph_pred)

	if train_mismatch and fit_type is "amp": #testing for amplitude
		PCA_train_ph = np.loadtxt(in_folder+"PCA_train_ph.dat")[:N_train,:]
		PCA_ph = PCA_model(in_folder+"ph_PCA_model")
		rec_ph=PCA_ph.reconstruct_data(PCA_train_ph)
		rec_amp=PCA.reconstruct_data(PCA_train)
		rec_amp_pred=PCA.reconstruct_data(PCA_train_pred)
		F_MoE_train = compute_mismatch(rec_amp, rec_ph, rec_amp_pred, rec_ph)

	if test_mismatch:
		print("Average MoE mismatch: ",np.mean(F_MoE))

	if test_mismatch and not train_mismatch:
		return np.mean(F_MoE), mse_test_list
	if train_mismatch and train_mismatch:
		return np.mean(F_MoE_train), np.mean(F_MoE), mse_train_list, mse_test_list
	return












"""
Module GW_generator.py
======================
	Definition of class MLGW_generator. The class generates a GW signal of a BBH coalescence when given orbital parameters of the BBH.
	Model performs the regression:
		theta = (q,s1,s2) ---> g ---> A, ph = W g
	First regression is done by a MoE model; the second regression is a PCA model. Some optional parameters can be given to specify the observer position.
	It makes use of modules EM_MoE.py and ML_routines.py for an implementation of a PCA model and a MoE fitted by EM algorithm.
"""
#################

import os
import sys
import warnings
import numpy as np
sys.path.insert(1, os.path.dirname(__file__)) 	#adding to path folder where mlgw package is installed (ugly?)
from EM_MoE import *			#MoE model
from ML_routines import *		#PCA model

#############DEBUG PROFILING
try:
	from line_profiler import LineProfiler

	def do_profile(follow=[]):
		def inner(func):
			def profiled_func(*args, **kwargs):
				try:
					profiler = LineProfiler()
					profiler.add_function(func)
					for f in follow:
						profiler.add_function(f)
					profiler.enable_by_count()
					return func(*args, **kwargs)
				finally:
					profiler.print_stats()
			return profiled_func
		return inner
except:
	pass

################# GW_generator class
class GW_generator(object):
	"""
GW_generator
============
	This class holds all the parts of ML models and acts as GW generator. Model is composed by a PCA model to reduce dimensionality of a WF datasets and by several MoE models to fit PCA in terms of source parameters. WFs can be generated both in time domain and frequency domain.
	Everything is hold in a PCA model (class PCA_model defined in ML_routines) and in two lists of MoE models (class MoE_model defined in EM_MoE). All models are loaded from files in a folder given by user. Files must be named exactly as follows:
		amp(ph)_exp_#		for amplitude (phase) of expert model for PCA component #
		amp(ph)_gat_#		for amplitude (phase) of gating function for PCA component #
		amp(ph)_feat		for list of features to use for MoE models
		amp(ph)_PCA_model	for PCA model for amplitude (phase)
		times/frequencies	file holding grid points at which waves generated by PCA are evaluated
	No suffixes shall be given to files.
	The class doesn't implement methods for fitting: it only provides a useful tool to gather them.
	"""
	def __init__(self, folder = "default"):
		"""
	__init__
	========
		Initialise class by loading models from file. "default" loads a pre-fitted default model (saved in "__dir__/TD_model").
		Everything useful for the model must be put within the folder with the standard names:
			{amp(ph)_exp_# ; amp(ph)_gat_#	; amp(ph)_feat ; amp(ph)_PCA_model; times/frequencies}
		There can be an arbitrary number of exp and gating functions as long as they match with each other and they are less than PCA components.
		A compulsory file times/frequencies must hold a list of grid points at which the generated ML wave is evaluated.
		Input:
			folder				address to folder in which everything is kept (if None, models must be loaded manually with load())
		"""
		self.times = None
		
		if folder is not None:
			if folder == "default":
				folder = os.path.dirname(__file__)+"/TD_model"
			self.load(folder)
		return

	def load(self, folder):
		"""
	load
	====
		Builds up all the models from given folder.
		Everything useful for the model must be put within the folder with the standard names:
			{amp(ph)_exp_# ; amp(ph)_gat_#	; amp(ph)_feat ; amp(ph)_PCA_model}
		There can be an arbitrary number of exp and gating functions as long as they match with each other and they are less than PCA components.
		It loads frequencies.
		Input:
			address to folder in which everything is kept
		"""
		if not folder.endswith('/'):
			folder = folder + "/"
		print("Loading model from: ", folder)
		file_list = os.listdir(folder)

			#loading PCA
		self.amp_PCA = PCA_model()
		self.amp_PCA.load_model(folder+"amp_PCA_model")
		self.ph_PCA = PCA_model()
		self.ph_PCA.load_model(folder+"ph_PCA_model")

		print("  Loaded PCA model for amplitude with ", self.amp_PCA.get_V_matrix().shape[1], " PC")
		print("  Loaded PCA model for phase with ", self.ph_PCA.get_V_matrix().shape[1], " PC")

			#loading features
		f = open(folder+"amp_feat", "r")
		self.amp_features = f.readlines()
		for i in range(len(self.amp_features)):
			self.amp_features[i] = self.amp_features[i].rstrip()

		f = open(folder+"ph_feat", "r")
		self.ph_features = f.readlines()
		for i in range(len(self.ph_features)):
			self.ph_features[i] = self.ph_features[i].rstrip()
		
		print("  Loaded features for amplitude: ", self.amp_features)
		print("  Loaded features for phase: ", self.ph_features)
	
			#loading MoE models
		print("  Loading MoE models")
			#amplitude
		self.MoE_models_amp = []
		k = 0
		while "amp_exp_"+str(k) in file_list and  "amp_gat_"+str(k) in file_list:
			self.MoE_models_amp.append(MoE_model(3+len(self.amp_features),1))
			self.MoE_models_amp[-1].load(folder+"amp_exp_"+str(k),folder+"amp_gat_"+str(k))
			print("    Loaded amplitude model for comp: ", k)
			k += 1
		
			#phase
		self.MoE_models_ph = []
		k = 0
		while "ph_exp_"+str(k) in file_list and  "ph_gat_"+str(k) in file_list:
			self.MoE_models_ph.append(MoE_model(3+len(self.ph_features),1))
			self.MoE_models_ph[-1].load(folder+"ph_exp_"+str(k),folder+"ph_gat_"+str(k))
			print("    Loaded phase model for comp: ", k)
			k += 1

		if "times" in file_list:
			print("  Loaded time vector")
			self.times = np.loadtxt(folder+"times")
		else:
			raise RuntimeError("Unable to load model: no time vector given!")

		np.matmul(np.zeros((2,2)),np.ones((2,2))) #this has something to do with a speed up of matmul. Once it is called once, matmul gets much faster!
		return

	def MoE_models(self, model_type, k_list=None):
		"""
	MoE_models
	==========
		Returns the MoE model(s).
		Input:
			model_type		"amp" or "ph" to state which MoE models shall be returned
			k_list []		index(indices) of the model to be returned (if None all models are returned)
		Output:
			models []	list of models to be returned
		"""
		if k_list is None:
			k_list = range(self.K)
		if model_type == "amp":
			return self.MoE_models_amp[k]
		if model_type == "ph":
			return self.MoE_models_ph[k]
		return None

	def PCA_models(self, model_type):
		"""
	PCA_models
	==========
		Returns the MoE model(s).
		Input:
			model_type		"amp" or "ph" to state which MoE models shall be returned
		Output:
			
		"""
		if model_type == "amp":
			return self.amp_PCA
		if model_type == "ph":
			return self.ph_PCA
		return None

	def model_summary(self, filename = None):
		"""
	PCA_models
	==========
		Prints to screen a summary of the model currently used.
		If filename is given, output is redirected to file.
		Input:
		Output:
			filename	if not None, redirects the output to file
		"""
		amp_exp_list = [str(model.get_iperparams()[1]) for model in self.MoE_models_amp]
		ph_exp_list = [str(model.get_iperparams()[1]) for model in self.MoE_models_ph]

		output = "###### Summary for MLGW model ######\n"
		output += "   Grid size:     "+str(self.amp_PCA.get_PCA_params()[0].shape[0]) +" \n"
		output += "   Minimum time:  "+str(np.abs(self.times[0]))+" s/M_sun\n"
			#amplitude summary
		output += "   ## Model for Amplitude \n"
		output += "      - #PCs:          "+str(self.amp_PCA.get_PCA_params()[0].shape[1])+"\n"
		output += "      - #Experts:      "+(" ".join(amp_exp_list))+"\n"
		output += "      - #Features:     "+str(self.MoE_models_amp[0].get_iperparams()[0])+"\n"
		output += "      - Features:      "+(" ".join(self.amp_features))+"\n"
			#phase summary
		output += "   ## Model for Phase \n"
		output += "      - #PCs:          "+str(self.ph_PCA.get_PCA_params()[0].shape[1])+"\n"
		output += "      - #Experts:      "+(" ".join(ph_exp_list))+"\n"
		output += "      - #Features:     "+str(self.MoE_models_ph[0].get_iperparams()[0])+"\n"
		output += "      - Features:      "+(" ".join(self.ph_features))+"\n"
		output += "####################################"
	
		if type(filename) is str:
			text_file = open(filename, "a")
			text_file.write(output)
			text_file.close()
			return
		elif filename is not None:
			warnings.warn("Filename must be a string! "+str(type(filename))+" given. Output is redirected to standard output." )
		print(output)
		return

	def get_time_grid(self):
		"""
	get_time_grid
	=============
		Returns the time grid at which the outputs of the models are evaluated. Grid is in reduced units.
		Output:
			time_grid (D,)	points in time grid at which all waves are evaluated
		"""
		return self.times


	def __call__(self, t_grid, m1, m2, spin1_x, spin1_y, spin1_z, spin2_x, spin2_y, spin2_z, D_L, i, phi_0, long_asc_nodes, eccentricity, mean_per_ano , plus_cross = True):
		"""
	__call__
	========
		Generates a WF according to the MLGW model. It makes all the required preprocessing to include wave dependance on the full 14 parameters space of the GW forms.
		Input:
			t_grid	(N_grid,)		Grid of (physical) time/frequency points to evaluate the wave at
			m1	()/(N,)				Mass of BH 1
			m2	()/(N,)				Mass of BH 1
			spin1_x/y/z	()/(N,)		Each variable represents a spin component of BH 1
			spin2_x/y/z				Each variable represents a spin component of BH 1
			D_L	()/(N,)				Luminosity distance
			i ()/(N,)				Inclination
			phi_0 ()/(N,)			Reference phase for the wave
			long_asc_nodes ()/(N,)	Logitudinal ascentional nodes (currently not implemented)
			eccentricity ()/(N,)	Eccentricity of the orbit (currently not implemented)
			mean_per_ano ()/(N,)	Mean per ano (currently not implemented)
			plus_cross				Whether to return h_+ and h_x components (if false amp and phase are returned)
		Ouput:
			h_plus, h_cross (1,D)/(N,D)	desidered polarizations (if it applies)
			amp, phase (1,D)/(N,D)		desidered amplitude and phase (if it applies)
		"""
		theta = np.column_stack((m1, m2, spin1_x, spin1_y, spin1_z, spin2_x, spin2_y, spin2_z, D_L, i, phi_0, long_asc_nodes, eccentricity, mean_per_ano)) #(N,D)
		return self.get_WF(theta, plus_cross = plus_cross, t_grid= t_grid, red_grid = False)


	def get_WF(self, theta, t_grid = None, plus_cross = True, red_grid = False):
		"""
	get_WF
	======
		Generates a WF according to the MLGW model. It makes all the required preprocessing to include wave dependance on the full 15 parameters space of the GW forms.
		Wherever not specified, all waves are evaluated at a luminosity distance of 1 Mpc.
		It accepts data in one of the following layout of D features:
			D = 3	[q, spin1_z, spin2_z]
			D = 4	[m1, m2, spin1_z, spin2_z]
			D = 5	[m1, m2, spin1_z , spin2_z, D_L]
			D = 6	[m1, m2, spin1_z , spin2_z, D_L, inclination]
			D = 7	[m1, m2, spin1_z , spin2_z, D_L, inclination, phi_0]
			D = 14	[m1, m2, spin1 (3,), spin2 (3,), D_L, inclination, phi_0, long_asc_nodes, eccentricity, mean_per_ano]
		Warning: last layout (D=14) is made only for compatibility with lalsuite software. The implemented variables are those in D=7 layout; the other are dummy variables and will not be considered.
		Unit of measures:
			[mass] = M_sun
			[D_L] = Mpc
			[spin] = adimensional
		Input:
			theta (N,D)		source parameters to make prediction at
			t_grid (D',)	a grid in (physical or reduced) time/frequency to evaluate the wave at (uses np.inter)
			plus_cross		whether to return h_+ and h_x components (if false amp and phase are returned)
			red_grid		whether given t_grid is in reduced space (True) or physical space (False)
		Ouput:
			h_plus, h_cross (N,D)	desidered polarizations (if it applies)
			amp,ph (N,D)			desidered amplitude and phase (if it applies)
		"""
		if t_grid is None:
			t_grid = self.times
			if red_grid == False:
				red_grid = True
				warnings.warn("As no grid is given, the default reduced grid is used to evaluate the output. red_grid option is set to True.")

		theta = np.array(theta) #to ensure user theta is copied into new array
		if theta.ndim == 1:
			theta = theta[np.newaxis,:] #(1,D)
		
		D= theta.shape[1] #number of features given
		if D <3:
			raise RuntimeError("Unable to generata WF. Too few parameters given!!")
			return

			#creating a standard theta vector for __get_WF__
		if D>3 and D!=7:
			new_theta = np.zeros((theta.shape[0],7))
			new_theta[:,4] = 1.
			if D== 14:
				if np.any(np.column_stack((theta[:,2:4], theta[:,5:7])) != 0):
					warnings.warn("Given nonzero spin_x/spin_y components. Model currently supports only spin_z component. Other spin components are ignored")
				indices = [0,1,4,7,8,9,10]
				indices_new_theta = range(7)
			else:
				indices = [i for i in range(D)]
				indices_new_theta = indices
		
				#building vector to keep standard layout for __get_WF__
			new_theta[:, indices_new_theta] = theta[:,indices]
			theta = new_theta #(N,7)


			#generating waves and returning to user
		res1, res2 = self.__get_WF__(theta, t_grid, red_grid, plus_cross)
			#res1,res2 = h_plus, h_cross if plus_cross = True
			#res1,res2 = amp, ph if plus_cross = False
		return res1, res2

	#@do_profile(follow=[])
	def __get_WF__(self, theta, time_grid, red_grid, plus_cross = False):
		"""
	__get_WF__
	==========
		Generates the waves in time domain and perform . Called by get_WF.
		Accepts only input features as [q,s1,s2] or [m1, m2, spin1_z , spin2_z, D_L, inclination, phi_0].
		Input:
			theta (N,D)		source parameters to make prediction at (D=3 or D=7)
			time_grid (D',)	a grid in (physical or reduced) time to evaluate the wave at (uses np.inter)
			red_grid		whether given t_grid is in reduced space (True) or physical space (False)
		Output:
			amp,ph (N,D)	desidered amplitude and phase
		"""
		D= theta.shape[1] #number of features given
		assert D in [3,7] #check that the number of dimension is fine

			#setting theta_std & m_tot_us
		if D == 3:
			theta_std = theta
			m_tot_us = 20. * np.ones((theta.shape[0],)) 
		else:
			q = np.divide(theta[:,0],theta[:,1]) #theta[:,0]/theta[:,1] #mass ratio (general) (N,)
			m_tot_us = theta[:,0] + theta[:,1]	#total mass in solar masses for the user
			theta_std = np.column_stack((q,theta[:,2],theta[:,3])) #(N,3)

			to_switch = np.where(theta_std[:,0] < 1.) #holds the indices of the events to swap

				#switching masses (where relevant)
			theta_std[to_switch,0] = np.power(theta_std[to_switch,0], -1)
			theta_std[to_switch,1], theta_std[to_switch,2] = theta_std[to_switch,2], theta_std[to_switch,1]

		amp, ph =  self.get_raw_WF(theta_std) #raw WF (N, N_grid)
		amp = 1e-21*amp #scaling back amplitude: less numerically efficient but quicker than doing it at the end

			#doing interpolations
		m_tot_std = 20.
			############
		new_amp = np.zeros((amp.shape[0], time_grid.shape[0]))
		new_ph = np.zeros((amp.shape[0], time_grid.shape[0]))

		for i in range(amp.shape[0]):
			if not red_grid:
				interp_grid = np.divide(time_grid,m_tot_us[i])
			else:
				interp_grid = time_grid
				#putting the wave on the user grid
			new_amp[i,:] = np.interp(interp_grid, self.times, amp[i,:]* m_tot_us[i]/m_tot_std ,left = 0, right = 0) #set to zero outside the domain
			new_ph[i,:]  = np.interp(interp_grid, self.times, ph[i,:])

				#warning if the model extrapolates outiside the grid
			if (interp_grid[0] < self.times[0]):
				warnings.warn("Warning: time grid given is too long for the fitted model. Set 0 amplitude outside the fitting domain.")

		amp = new_amp 
		ph = np.subtract(new_ph.T,new_ph[:,0]).T #phase are zero at t = 0 #SLOOOW: do you need it??

			#### Dealing with distance, inclination and phi_0
		if D==7: #distance corrections are done
			dist_pref = theta[:,4] #std_dist = 1 Mpc
			iota = theta[:,5] #std_inclination = 0.
			phi_0 = theta[:,6] #reference phase

				#scaling to required distance
			#amp = np.divide(amp.T, dist_pref).T #slow... check better if it works...

			#scaling for setting inclination (it is done only if required)
				#checking whether workin with real things is better
			h_22_real = np.multiply(amp, np.cos(ph) )
			h_22_imag = np.multiply(amp, np.sin(ph) )
			#h_22 = h_22_real + 1j*h_22_imag 
			#h_22 = np.multiply(amp, np.exp(1j*(ph)) ) #choose here a convention... (lal is +) #very slow!!!!

					#parametrization of the wave
				#h = h_p +i h_c = Y_22 * h_22 + Y_2-2 * h_2-2
				#h_22 = h*_2-2
				#Y_2+-2 = sqrt(5/(64pi))*(1+-cos(inclination))**2 exp(+-2i phi)

			#h = np.multiply(h_22.T, self.__Y_2m__(2,iota, phi_0)).T + np.multiply(np.conj(h_22).T, self.__Y_2m__(-2,iota, phi_0)).T #very slow to work with complex numbers 
			h_p, h_c = self.__set_d_iota_phi_dependence__(h_22_real,h_22_imag, dist_pref, iota, phi_0)

			if plus_cross:
				return h_p, h_c
			else: 
				amp =  np.sqrt(np.square(h_p)+np.square(h_c)) 
				ph = np.unwrap(np.arctan2(h_c,h_c))
				return amp, ph

		if plus_cross:
			h_p = np.multiply(amp, np.cos(ph))
			h_c = np.multiply(amp, np.sin(ph))
			return h_p, h_c
		else:
			return amp, ph				

	def __set_d_iota_phi_dependence__(self, h_p, h_c, dist, iota, phi_0):
		"""
	__set_d_iota_phi_dependence__
	=============================
		DO IT!!!!!!!!!!!!!!!!
		Input:
			h_p, h_c (N,D)
			iota (N,)
			phi_0 (N,)
		"""
		const = 1./4.
		c_i = np.cos(iota) #(N,)
			#dealing with h_p
		new_h_p = np.multiply(h_p.T, np.cos(2*phi_0)) - np.multiply(h_c.T, np.sin(2*phi_0)) #(D,N) #included phi dependence
		new_h_p = np.multiply(new_h_p, 2*const*(1+np.square(c_i))/dist ).T #(N,D) #included iota dependence

			#dealing with h_p
		new_h_c = np.multiply(h_p.T, np.sin(2*phi_0)) + np.multiply(h_c.T, np.cos(2*phi_0)) #(D,N) #included phi dependence
		new_h_c = np.multiply(new_h_c, 4*const*c_i/dist ).T #(N,D) #included iota dependence

		return new_h_p, new_h_c

	def __Y_2m__(self,m, iota, phi):
		"""
	__Y_2m__
	========
		Spin 2 spherical harmonics with l = 2. Only m = +2,-2 is implemented.
		It has the following expression:
			Y_l+-2 = (1+-cos(iota))**2 e**(2i*m*phi_0)
		Input:
			m			index m (only +/-2 is implemented)
			iota (N,)	inclination angle
			phi_0 (N,)	reference phase
		Output:
			Y_2m (N,)	value of the function
		"""
		const = 1./4.#np.sqrt(5./(64*np.pi)) #the constant was already fitted in the wave
		c_i = np.cos(iota) #(N,)
		Y_2m = const * np.square(1+np.multiply(np.sign(m), c_i)) #(N,)
		Y_2m = np.multiply(Y_2m, np.exp(1j*np.multiply(np.sign(m), 2*phi)) ) #(N,)
		return Y_2m

	def get_raw_WF(self, theta):
		"""
	get_raw_WF
	==========
		Generates a WF according to the MLGW model with a parameters vector in MLGW model style (params=  [q,s1z,s2z]).
		All waves are evaluated at a luminosity distance of 1 Mpc and inclination 0. They are generated at masses m1 = q * m2 and m2 = 20/(1+q), so that M_tot = 20.
		Grid is the standard one.
		Input:
			theta (N,3)		source parameters to make prediction at
		Ouput:
			amp,ph (N,D)	desidered amplitude and phase
		"""
		rec_PCA_dataset_amp, rec_PCA_dataset_ph = self.get_red_coefficients(theta)

		rec_amp_dataset = self.amp_PCA.reconstruct_data(rec_PCA_dataset_amp)
		rec_ph_dataset = self.ph_PCA.reconstruct_data(rec_PCA_dataset_ph)

		return rec_amp_dataset, rec_ph_dataset

	#@do_profile(follow=[])
	def get_red_coefficients(self, theta):
		"""
	get_red_coefficients
	====================
		Returns the PCA reduced coefficients, as estimated by the MoE models.
		Input:
			theta (N,3)		source parameters to make prediction at
		Ouput:
			red_amp,red_ph (N,K)	PCA reduced amplitude and phase
		"""
		assert theta.shape[1] == 3

			#adding extra features
		amp_theta = add_extra_features(theta, self.amp_features)
		ph_theta = add_extra_features(theta, self.ph_features)

			#making predictions for amplitude
		rec_PCA_dataset_amp = np.zeros((amp_theta.shape[0], self.amp_PCA.get_dimensions()[1]))
		for k in range(len(self.MoE_models_amp)):
			rec_PCA_dataset_amp[:,k] = self.MoE_models_amp[k].predict(amp_theta) #why is this much sower than the phase part?

			#making predictions for phase
		rec_PCA_dataset_ph = np.zeros((ph_theta.shape[0], self.ph_PCA.get_dimensions()[1]))
		for k in range(len(self.MoE_models_ph)):
			rec_PCA_dataset_ph[:,k] = self.MoE_models_ph[k].predict(ph_theta)

		return rec_PCA_dataset_amp, rec_PCA_dataset_ph

########THIS IS SHIIIIT!!########
	def align_wave(self, amp, ph, t_grid = None, al_merger = True, phi_0=0):
		"""
	align_wave
	==========
		Given a set of waves in time domain, it sets time scale s.t. amplitude is max at t=0 (if a grid is given)
		It sets ph = ph_0 at t=0 if al_merger is true; ph=ph_0 at beginning of time grid if al_merger is False.
		Input:
			amp	(N, N_grid)	amplitude of waves
			ph (N,N_grid)	phases of waves
			t_grid (N_grid)	grid at which wave is evaluated	
			phi_0 ()/(N,)	value of phase at the point at which is aligned (default =0)
		Output:
			amp, ph (N,N_grid)	amplitude and phases rescaled appropriately
		"""
		assert amp.shape == ph.shape
		if amp.ndim == 1:
			amp = amp[None,:]
			ph = ph[None,:]

		if t_grid is not None: #adjusting amplitude
			argmax_amp = np.argmax(amp, axis = 1) #(N,)
			#huge_grid = np.linspace(t_grid[0], t_grid[-1], int(1e5))
			for i in range(amp.shape[0]):
				amp[i,:] = np.interp( t_grid, t_grid- t_grid[argmax_amp[i]], amp[i,:])
				ph[i,:]  = np.interp( t_grid, t_grid- t_grid[argmax_amp[i]], ph[i,:])
		
			#aligning phase
		if not al_merger:
			return amp, (np.subtract(ph.T,ph[:,0]) + phi_0).T #this works a lot but is very unelegant!!!

		if al_merger:
			for i in range(amp.shape[0]):
				ph_merger = ph[i,np.argmax(amp[i,:])]
				ph[i,:] = ph[i,:] - ph_merger
			ph = (ph.T + phi_0).T
		return amp, ph




"""
mlgw
====
Package for a ML model to generate a GW signal by a BBH coalescence. The model makes use of a PCA model for reducing wave dimensionality. A Mixture of Expert model makes a regression from orbital parameters to the reduced representation of a wave.
The packacke has 4 modules:
	GW_generator.py
		It holds an implementation of the full ML model used to reconstruct a WF. It is a lal style method for creating a waveform. It is ready to use. Loads authomatically EM_MoE and ML_routines modules.
	EM_MoE.py
		It holds a Mixture of Experts model as well as a softmax regression model required for it.
	ML_routines.py
		It holds some useful ML routines such as PCA model (required by the MLGW_generator) and a routine for enlarging a dataset dimensionality.
	GW_helper.py
		It holds some routines useful for generating a GW dataset and a computing mismatch between waveforms. This is not strictly required by the model but it is useful for training the model. It is used by shared code for fitting and created dataset.
		
"""



"""
Module ML_routines.py
=====================
	Definition of the following ML routines:
		PCA model
			class PCA_model: implements a PCA model with methods for fitting and doing data reduction
		Gaussian Discriminant Analysis
			class GDA: implements a model for a Gaussian discriminant Analysis classifiers. It might be useful for MoE.
		Data augmentation helper
			function add_extra_features: adds to a dataset some extra polynomial features
"""
#################

import scipy.stats
import numpy as np
import warnings

################# PCA class
class PCA_model:
	"""
PCA_model
=========
	Class aimed to deal with a PCA model.
	It fits a PCA model and is able to reduce a dataset (dimension D) to a lower dimensional (dimension K) one and to reconstruct low dimensional data to high dimensional one.
	It stores the following parameters (get them with get_PCA_params()):
		V (D,K)			matrix for dimensional reduction
		mu (D,)			the average value for each feature of dataset
		max_PC (K,)		maximum value of PC projection used to redurn scaled low dimensional data (activate it with scale_PC=True in fit_model methods())
		E (K,)			Eigenvalues of the PCs
	"""
	def __init__(self, filename = None):
		"""
	__init__
	========
		Constructor for PCA model. If filename is given, loads the model from file.
		Input:
			filename	file to load the model from
		"""
		self.PCA_params = []
		if filename is not None:
			self.load_model(filename)
		return None

	def save_model(self, filename):
		"""
	save_model
	==========
		Save the PCA model parameters to file in the matrix
			[[V (D,K), mu (D,)], [max_PC (K(+1),)], [E (K(+1),)] ]
		with shape (D+2,K+1)
		Input:
			filename	file to save the model in
		Output:
		"""
		if self.PCA_params == []:
			print("Model is not fitted yet! There is nothing to save")
			return None
		(D, K) = self.PCA_params[0].shape
		V = self.PCA_params[0] #(D,K)
		mu = (self.PCA_params[1])[:,np.newaxis]#(D,1)
		max_PC = (self.PCA_params[2]) #(K,)
		E = (self.PCA_params[3]) #(K,)
		first_row = np.concatenate((V,mu), axis = 1) #(D, K+1)

		to_save = np.zeros((D+2, K+1))
		to_save[:D,:] = first_row
		to_save[D,:K] = max_PC
		to_save[D+1,:K] = E.real
		to_save[D:,-1] = np.NAN		

		np.savetxt(filename, to_save)
		return None 

	def load_model(self, filename):
		"""
	load_model
	==========
		Load the PCA parameters from file. The format is the same as save_model
		Input:
			filename	file to load the model from
		Output:
		"""
		data = np.loadtxt(filename) #loading data

		if not np.any(np.isnan(data)): #if there is no NaN, the old format is employed. This is to ensure code portability :(
			warnings.warn("Old PCA model type given. The model is loaded correctly but it is better to save the model to the new format.")
			V = data[:,:data.shape[1]-2] #(D,K)
			mu = data[:,data.shape[1]-2] #(D,)
			max_PC = data[:V.shape[1],data.shape[1]-1] #(K,)
			E = np.ones((V.shape[1],)) #(K,)
		else:
			V = data[:data.shape[0]-2,:data.shape[1]-1]
			mu = data[:data.shape[0]-2,data.shape[1]-1]
			max_PC = data[data.shape[0]-2,:data.shape[1]-1]
			E = data[data.shape[0]-1,:data.shape[1]-1]

		self.PCA_params= [V,mu,max_PC, E]
		return None

	def reconstruct_data(self, red_data):
		"""
	reconstruct_data
	================
		Gives the best estimate of high dimensional data given the low dimensional PCA approximation.
		Data are rescaled back to the original training measure inverting the preprocessing procedure.
		Input:
			red_data (N,K)	low dimensional representation of data
		Output:
			data (N,D)		high dimensional reconstruction of data (after inversion of preprocessing)
		"""
		red_data = np.multiply(red_data, self.PCA_params[2])
		data = np.matmul(red_data, self.PCA_params[0].T)
		data = data+self.PCA_params[1]
		return data.real


	def reduce_data(self, data):
		"""
	reduce_data
	===========
		Reduce data by applying a PCA dimensionality reduction. Data are preprocessed by a scaling and a mean shift according to parameters give in PCA_params (in the fashion of fit_PCA). The reduced version of preprocessd data is returned (X_red = X*V_PCA).
		Input:
			data (N,D)		data to reduce
		Output:
			red_data (N,K)	dimensional reduction of preprocessed data
		"""
		data = data - self.PCA_params[1]
		red_data = np.matmul(data, self.PCA_params[0])
		red_data = np.divide(red_data, self.PCA_params[2]) #scaling PC to make them o(1)
		return red_data.real

	def fit_model(self, X, K = None, scale_PC=True):
		"""
	fit_model
	=========
		Fit the PCA model for the given dataset. Data are done zero mean for each feature and rescaled s.t. are O(1) if scale_data is True.
		A parameter set is returned holding fitted PCA parameters (projection matrix, data mean and scale factor)
		Input:
			X (N,D)		training set
			K ()		number of principal components
			scale_PC	whether PC should be scaled by their maximum value to make them all O(1)
		Output:
			E (K,)	eigenvalues of the first K principal components
		"""
		X = X.real
		mu = np.mean(X,0) #(D,)
		X = X - mu

			#doing actual PCA
		if K is None:
			K = X.shape[1]
		#E, V = np.linalg.eig(np.dot(X.T, X))
		E, V = np.linalg.eig(np.cov(X.T))
		idx = np.argsort(E)[::-1]
		V = V[:, idx[:K]] # (D,K)
		self.PCA_params = [V.real, mu, np.ones((K,)),  E[:K].real]

		if scale_PC:
			red_data = np.matmul(X, self.PCA_params[0]) #(N,K)
			self.PCA_params[2] = np.max(np.abs(red_data), axis = 0) #(K,)

		return E[:K].real

	def get_V_matrix(self):
		"""
	get_V_matrix
	============
		Returns the projection matrix of the model
		Input:
		Output:
			V (D,K) matrix of eigenvector used for projection and reconstruction of data
		"""
		return self.PCA_params[0]

	def get_dimensions(self):
		"""
	get_dimensions
	==============
		Returns dimension of high- and low-dimensional space.
		Input:
		Output:
			(D,K) (tuple)	dimensions in the format (high-dim, low-dim)
		"""
		return self.PCA_params[0].shape

	def get_PCA_params(self):
		"""
	get_PCA_params
	==============
		Returns the parameters of the model
		Input:
		Output:
			PCA_params [V (D,K), mu (D,), max_PC (K,), E (K,)]	paramters for preprocessing and PCA
		"""
		return self.PCA_params

################# Gaussian Discriminant Analysis
class GDA(object):
	"""
GDA
===
	This class implements a model for Gaussian Discriminant Analysis. The model is a classifier with form:
		p(y=k|x,params) ~ p(y=k) * p(x | y=k, params) =  pi_k * N(x | mu_k, sigma_k)
	"""
	def __init__(self, D, K, naive = True, hard_clustering = False, same_weights = False):
		"""
	__init__
	========
		Initialize the model with K classes for regressions.
		Input:
			D 				dimensionality of input space
			K				number or classes for regression
			naive			whether the covariance matrix for each class should be diagonal (naive assumption)
			hard_clustering	whether hard clustering predictions should be done
			same_weights	whethet all p(y=k) should be the same
		"""
		self.D = D 
		self.K = K 
		self.naive = naive
		self.same_weights = same_weights
		self.hard_clustering = hard_clustering
		if not naive: self.models = []
		self.model_params = []
			#initializing with dummy things
		for k in range(self.K):	
			mu = np.zeros((D,)) #(D,)
			sigma = np.ones((D,))/D
			if not naive:
				self.models.append( scipy.stats.multivariate_normal(mu, np.diag(sigma)) )
				self.model_params.append((mu,np.diag(sigma)))
			else:
				self.model_params.append((mu,sigma))
		self.pi_k = np.ones((K,))/K #initialization
		return

	def init_centroids(self, centroids, sigma = None):
		"""
	init_centroids
	==============
		Initialize each of cluster mean and variance with a custom value..
		Input:
			centroids (D,K)		array with K centroids (each of D dimension)
			sigma (D,K)			diagonal of covariance matrix
		"""
		if centroids.shape != (self.D,self.K):
			raise TypeError("Centroids matrix not suitable! Shape should be ("+str(self.D)+","+str(self.K)+")")
		if sigma is None:
			sigma = np.ones((self.D,self.K))
		if sigma.shape != (self.D,self.K):
			raise TypeError("Sigma matrix not suitable!")
		for k in range(self.K):
			self.model_params[k] = (centroids[:,k],sigma[:,k])
		return 

	def predict(self, X_test, LL = False):
		"""
	predict
	=======
		Makes predictions with probability: p(y = k | x).
		Log probability can be returned
		Input:
			X_test (N,D)	test points
			LL				whether log probability is required
		Ouput:
			y_pred (N,K)	predictions
		"""
		if len(self.model_params) != self.K:
			raise RuntimeWarning("Model must be fitted before making predictions! No predictions have been done")
			return None
		res = np.zeros((X_test.shape[0],self.K))		
		for k in range(self.K):
			if self.naive:
				mu = self.model_params[k][0]	#(D,)
				sigma_sq = self.model_params[k][1] #(D,) diagonal elements
				non_zero = np.where(sigma_sq !=0)[0] #(D',) only non-zero variance components enter the valuation of argument
				arg = -0.5*np.divide(np.square(X_test[:,non_zero]- mu[non_zero]), sigma_sq[non_zero]) #shape (N,D')
				arg = np.sum(arg, axis =1) #(N,)
				#arg = arg - .5 * np.sum(np.log(sigma_sq[non_zero])) #it's better not to normalize things
				res[:,k] = arg + np.log(self.pi_k[k]) #adding p(y=k) as in Bayes' rule
			if not self.naive:
				res[:,k] = self.pi_k[k]*self.models[k].pdf(X_test) #(N,) #really old version of the model

		if self.hard_clustering: #returning hard clustering predictions
			to_return = np.zeros(res.shape)
			to_return[np.arange(len(res)), res.argmax(1)] = 1
			return to_return

		if not self.naive:
			res = np.divide(res.T, np.sum(res, axis =1)).T
			if LL: res = np.log(res)
			return res #(N,K)	
		else:
			if LL:
				return res #result is not normalized... I should find a way to do it...
			res = (res.T - np.mean(res, axis =1)).T
			res = np.exp(res)
			res[np.where(res==0)[0]] = 1e-5
			return np.divide(res.T, np.sum(res, axis =1)).T

	def fit(self, X_train, y_train):
		"""
	fit
	===
		Fit the model with MLE.
		Input:
			X_train (N,D)	train data
			y_train (N,)	train labels
		"""
		self.models = []
		self.model_params = []
		for k in range(self.K):	
			mu = np.sum(np.multiply(X_train.T,y_train[:,k]).T, axis = 0) #(D,)
			mu = np.divide(mu, np.sum(y_train[:,k]))
			#print("mu:", mu.shape)
			temp_var = X_train-mu
			#print("temp_var: ", temp_var.shape)
			sigma_sq = np.matmul(np.multiply((X_train-mu).T, y_train[:,k]), X_train-mu) #(D,D)
			sigma_sq = np.divide(sigma_sq, np.sum(y_train[:,k]))
			if not self.naive:
				self.models.append( scipy.stats.multivariate_normal(mu, sigma_sq) )
			if self.naive:
				#only elements in diagonal are taken (naive assumption)
				sigma_sq = np.diag(sigma_sq) #(D,) #can be used to prevent overfitting
			self.model_params.append((mu,sigma_sq))
		if not self.same_weights:
			self.pi_k = np.sum(y_train, axis = 0)/np.sum(y_train)
		return
	
	def get_weights(self):
		"""
	get_weights
	===========
		Return the weights of the model.
		Output:
			model_params []		parameters of generative gaussians [(mu_0,sigma_0), ... , (mu_K-1,sigma_K-1)]
			pi_k (k,)			probabilities for each class p(y = k) 
		"""
		return self.model_params, self.pi_k

	def accuracy(self, X_test, y_test, LL = False):
		"""
	accuracy
	========
		Computes the accuracy of the model (i.e. the fraction of misclassified points).
		This measure is meaningful only in the case of hard clustering where there is only one label for each data point.
		Input:
			X_test (N,D)	test points
			y_test (N,K)	true labels of test points
			LL				whether to use log prob for predictions
		Output:
			accuracy	accuracy of the predictions made at test points
		"""
		if X_test.ndim == 1:
			X_test = np.reshape(X_test, (X_test.shape[0],1))
		y_pred = self.predict(X_test, LL)
		return np.sum(np.argmax(y_pred,axis=1)==np.argmax(y_test,axis=1))/float(y_test.shape[0])


################# Extra features routine
def add_extra_features(data, feature_list, scaling = None):
	"""
add_extra_features
==================
	Given a dataset, it enlarge its feature number to make a basis function regression.
	Features to add must be specified with feature list. Each element in the list is a string in form "ijk" where ijk are feature indices as in numpy array data (repetitions allowed); this represents the new feauture x_new = x_i*x_j*x_k
	Features can be scaled by an arbitrary constant value.
	Input:
		data (N,D)/(N,)			data to augment
		feature_list (len L)	list of features to add
		scaling (D,)			vector of values for scaling data as data_ij = data_ij/scaling_j (if None no scaling is performed)
	Output:
		new_data (N,D+L)	data with new feature
	"""
	data = np.array(data) #this is required to manipulate freely the data...
	if data.ndim == 1: data = data[:,np.newaxis]
	if len(feature_list)==0:
		return data

	data[:,0] = np.log(data[:,0]) #probably a good idea...

	#if scaling is not None:
	#	data = np.divide(data,scaling)

	new_features = np.zeros((data.shape[0],len(feature_list)))
	for i in range(len(feature_list)):
		feat_str = feature_list[i]
		temp = np.ones((data.shape[0],)) #(N,)
		if len(feat_str) ==1: continue #single feature is already in the data
		for k in feat_str:
			try:
				temp = np.multiply(temp, data[:,int(k)]) #(N,)
			except:
				raise RuntimeWarning("Unproper feature code for feature \""+k+"\": unable to add any the feature. Constant feature is placed instead")
				temp = np.ones((data.shape[0],))
				break
		new_features[:,i] = temp #assign new feature

	new_data = np.concatenate((data, new_features), axis = 1)
	return new_data





"""
Module GW_helper.py
===================
	Various routines for performing some tasks regarding GW signals (it uses lal and lalsimulation to work):
		Mismatch computation
			function compute_mismatch: computes the mismatch between two sets of wave (useful for seeing how important is reconstruction error)
		Scalar product computation
			function compute_scalar: computes the Wigner scalar product between two GW waveforms
		Dataset creation Time Domain
			function create_dataset_TD: creates a dataset of GW in time domain.
		Dataset creation Frequency Domain
			function create_dataset_FD: creates a dataset of GW in frequency domain.
"""
#################

import numpy as np
import lalsimulation as lalsim
import lal
import os.path
import matplotlib.pyplot as plt #debug

################# Overlap related stuff
def get_low_high_freq_index(flow, fhigh, df):
	kmin = int(flow / df)
	kmax = int(fhigh / df)
	return kmin, kmax

def overlap(amp_1, ph_1, amp_2, ph_2, df, low_freq = None, high_freq = None, PSD = None):
		#it works only for input shapes (D,)
	w1 = amp_1*np.exp(1j*ph_1)
	w2 = amp_2*np.exp(1j*ph_2)
	
	if PSD is None:
		PSD = np.ones((amp_1.shape[0],))
	if low_freq is not None and high_freq is not None:
		kmin, kmax = get_low_high_freq_index(low_freq, high_freq, df)
	else:
		kmin =0
		kmax = w1.shape[0]
	w1 = w1[kmin:kmax]
	w2 = w2[kmin:kmax]
	
	overlap = 2.0*df*np.sum(np.divide(np.conj(w1)*w2+np.conj(w2)*w1,PSD)).real
#	overlap = 4.0*df*np.sum(np.divide(np.conj(w1)*w2,PSD)).real
	return overlap

def compute_scalar(amp_1, ph_1, amp_2, ph_2, df, S = None):
	"""
compute_scalar
==============
	Computes an approximation to the Wigner scalar product for the two given GW.
		<h_1,h_2> = 4 Re[integral df h_1*(f) h_2(f)/S(f) ] ~ 4 Re sum[amp_1*(f_i) amp_2(f_i) exp(i*(ph_2(f_i)-ph_1(f_i)))/S(f_i) ] / D
	Input:
		amp_1/ph_1 (N,D)	amplitude/phase vector for wave 1 in Fourier space sampled in D uniform points within the domain
		amp_2/ph_2 (N,D)	amplitude/phase vector for wave 2 in Fourier space sampled in D uniform points within the domain
		df	()/(N,)			distance in the domain between two sampled points (can be uniform or different for each data point)
		S (N,D)				noise power spectral density sampled in D uniform points within the domain (if None there is no noise)
	Output:
		scalar (N,)	Wigner scalar product for the two waves
	"""
		#getting waves
	w1 = np.multiply(amp_1,np.exp(1j*ph_1))
	w2 = np.multiply(amp_2,np.exp(1j*ph_2))

		#checking for noise
	if S is None:
		S = np.ones(w1.shape)
	else:
		if S.shape != w1.shape:
			raise TypeError('Noise doesn\'t have the shape of data ('+str(amp_1.shape)+')')
			return None

	if w1.ndim < 2 and w2.ndim < 2:
		product = 4.0*np.multiply(df,np.sum(np.divide(np.conj(w1)*w2,S))).real
	else:	
		product = 4.0*np.multiply(df,np.sum(np.divide(np.conj(w1)*w2,S), axis = 1)).real
	
	return product.real


def compute_mismatch(amp_1, ph_1, amp_2, ph_2, S = None):
	"""
compute_mismatch
================
	Compute mismatch F between the waves given in input. Mismatch is computed with the formula
		F = 1-<h_1,h_2>/sqrt(<h_1,h_1><h_2,h_2>)
	with <,> being the Wigner scalar product for GW.
	Warning: waves must be aligned. Please use compute_optimal_mismatch for unaligned waves.
	Input:
		amp_1/ph_1 (N,D)	amplitude/phase vector for wave 1 in Fourier space sampled in D uniform points within the domain
		amp_2/ph_2 (N,D)	amplitude/phase vector for wave 1 in Fourier space sampled in D uniform points within the domain
		S (D,)				noise power spectral density sampled in D uniform points within the domain (if None there is no noise)
	Output:
		F (N,)	Mismatch between waves computed element-wise (i.e. F[i] holds mismatch between h_1[i,:] and h_2[i,:])
	"""
		#checking if data make sense
	if len({amp_1.shape, ph_1.shape, amp_2.shape, ph_2.shape}) == 1: #python set contains unique values
		if amp_1.ndim == 1:
			amp_1 = np.reshape(amp_1, (1,amp_1.shape[0]))
			amp_2 = np.reshape(amp_2, (1,amp_2.shape[0]))
			ph_1 = np.reshape(ph_1, (1,ph_1.shape[0]))
			ph_2 = np.reshape(ph_2, (1,ph_2.shape[0]))
		D = amp_1.shape[1]
		N = amp_1.shape[0]
		#print(N,D)
	else:
		pass
		#raise TypeError('Data don\'t have the same shape')
		#return None

		#computing mismatch vector
	df = 1. #nothing depends on df so we can set it arbitrarly
	F = compute_scalar(amp_1, ph_1, amp_2, ph_2, df, S)
	div_factor = np.sqrt(np.multiply(compute_scalar(amp_2, ph_2, amp_2, ph_2, df, S), compute_scalar(amp_1, ph_1, amp_1, ph_1, df, S)))
	np.divide(F, div_factor, out = F)
	return 1-F

def compute_optimal_mismatch(h1,h2, return_F = True):
	"""
compute_optimal_mismatch
========================
	Computes the optimal mismatch/overlap between two complex waveforms by performing the minimization:
		F = min_phi F[h1, h2*exp(1j*phi)]
	After the computation, h1 and h2*exp(1j*phi) are optimally aligned.
	Input:
		h1 (N,D)/(D,)	complex wave
		h2 (N,D)/(D,)	complex wave
		return_F		whether to reteurn mismatch or overlap
	Output:
		F_optimal (N,)/()		optimal mismatch/overlap
		phi_optimal (N,)/()		optimal phi which aligns the two waves
	"""
	assert h1.shape == h2.shape
	if h1.ndim == 1:
		h1 = h1[np.newaxis,:] #(1,D)
		h2 = h2[np.newaxis,:] #(1,D)

	scalar = lambda h1_, h2_: np.sum(np.multiply(h1_,np.conj(h2_)), axis = 1)/h1_.shape[1] #(N,)

	norm_factor = np.sqrt(np.multiply(scalar(h1,h1).real, scalar(h2,h2).real))
	overlap = scalar(h1,h2) #(N,)
	phi_optimal = np.angle(overlap) #(N,)
	overlap = np.divide(scalar(h1,h2*np.exp(1j*phi_optimal)), norm_factor)
	overlap = overlap.real

	#plt.plot(np.linspace(0,h1.shape[1],h1.shape[1]),np.squeeze(h1))
	#plt.plot(np.linspace(0,h1.shape[1],h1.shape[1]),np.squeeze(h2)*np.exp(1j*phi_optimal))
	#print(1-overlap, overlap)
	#plt.show()


	if return_F:
		return 1-overlap, phi_optimal
	if not return_F:
		return overlap, phi_optimal



################# Dataset related stuff

def create_dataset_TD(N_data, N_grid, filename = None,  t_coal = 0.5, q_range = (1.,5.), m2_range = None, s1_range = (-0.8,0.8), s2_range = (-0.8,0.8), t_step = 1e-5, lal_approximant = "SEOBNRv2_opt", alpha = 0.35):
	"""
create_dataset_TD
=================
	Create a dataset for training a ML model to fit GW waveforms in time domain.
	The dataset consists in 3 parameters theta=(q, spin1z, spin2z) associated to the waveform computed in frequency domain for a grid of N_grid points in the range given by the user.
	More specifically, data are stored in 3 vectors:
		theta_vector	vector holding source parameters q, spin1, spin2
		amp_vector		vector holding amplitudes for each source evaluated at some N_grid equally spaced points
		ph_vector		vector holding phase for each source evaluated at some N_grid equally spaced points
	This routine add N_data data to filename if one is specified (if file is not empty it must contain data with the same N_grid); otherwise the datasets are returned as np vectors. 
	All the waves are evaluated at a constant distance of 1Mpc. Values of q and m2 as well as spins are drawn randomly in the range given by the user: it holds m1 = q *m2 M_sun.
	The waveforms are computed with a time step t_step; starting from a frequency f_min (set by the routine according to t_coal and m_tot). Waves are given in a rescaled time grid (i.e. t/m_tot) with N_grid points: t=0 occurs when at time of maximum amplitude. A higher density of grid points is placed in the post merger phase.
	Dataset can be loaded with load_dataset.
	Input:
		N_data				size of dataset
		N_grid				number of grid points to evaluate
		filename			name of the file to save dataset in (If is None, nothing is saved on a file)
		t_coal				time to coalescence to start computation from (measured in reduced grid)
		q_range				tuple with range for random q values. if single value, q is kept fixed at that value
		m2_range			tuple with range for random m2 values. if single value, m2 is kept fixed at that value. If None, m2 will be chosen s.t. m_tot = m1+m2 = 20. M_sun
		spin_mag_max_1		tuple with range for random spin #1 values. if single value, s1 is kept fixed at that value
		spin_mag_max_2		tuple with range for random spin #1 values. if single value, s2 is kept fixed at that value
		t_step				time step to generate the wave with
		lal_approximant		string for the approximant model to be used (in lal convention)
		alpha				distorsion factor for time grid. (In range (0,1], when it's close to 0, more grid points are around merger)
	Output:
		if filename is given
			None
		if filename is not given
			theta_vector (N_data,3)		vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
			amp_dataset (N_data,N_grid)	dataset with amplitudes
			ph_dataset (N_data,N_grid)	dataset with phases
			times (N_grid,)				vector holding times at which waves are evaluated (t=0 is the time of maximum amplitude)
	"""
	d=1.
	inclination = 0.#np.pi/2.
	if d != 1. or inclination != 0.: #debug
		print("d is not standard!!! ",d) #debug
		print("i is not standard!!! ",inclination) #debug
	LALpars = lal.CreateDict()
	approx = lalsim.SimInspiralGetApproximantFromString(lal_approximant)

		#checking if N_grid is fine
	if not isinstance(N_grid, int):
		raise TypeError("N_grid is "+str(type(N_grid))+"! Expected to be a int.")

	if isinstance(m2_range, tuple):
		D_theta = 4 #m2 must be included as a feature
	else:
		D_theta = 3

		#creating time_grid
	t_end = 5.2e-4 #estimated maximum time for ringdown: WF will be killed after that time
	#alpha = 0.35 #exponent for "time distorsion"
	time_grid = np.linspace(-np.power(np.abs(t_coal), alpha), np.power(t_end, alpha), N_grid)
	time_grid = np.multiply( np.sign(time_grid) , np.power(np.abs(time_grid), 1./alpha))

		#adding 0 to time grid
	index_0 = np.argmin(np.abs(time_grid))
	time_grid[index_0] = 0. #0 is alway set in the grid

		#setting t_coal_freq for generating a waves
	if np.abs(t_coal) < 0.05:
		t_coal_freq = 0.05
	else:
		t_coal_freq = np.abs(t_coal)


	if filename is not None: #doing header if file is empty - nothing otherwise
		if not os.path.isfile(filename): #file doesn't exist: must be created with proper header
			filebuff = open(filename,'w')
			print("New file ", filename, " created")
			freq_header = np.concatenate((np.zeros((3,)), time_grid, time_grid) )
			freq_header = np.reshape(freq_header, (1,len(freq_header)))
			np.savetxt(filebuff, freq_header, header = "row: theta "+str(D_theta)+" | amp (1,"+str(N_grid)+")| ph (1,"+str(N_grid)+")\nN_grid = "+str(N_grid)+" | t_coal ="+str(t_coal)+" | t_step ="+str(t_step)+" | q_range = "+str(q_range)+" | m2_range = "+str(m2_range)+" | s1_range = "+str(s1_range)+" | s2_range = "+str(s2_range), newline = '\n')
		else:
			filebuff = open(filename,'a')

	if filename is None:
		amp_dataset = np.zeros((N_data,N_grid)) #allocating storage for returning data
		ph_dataset = np.zeros((N_data,N_grid))
		theta_vector = np.zeros((N_data,D_theta))

	for i in range(N_data): #loop on data to be created
		if i%50 == 0 and i != 0:
			print("Generated WF ", i)

			#setting value for data
		if isinstance(m2_range, tuple):
			m2 = np.random.uniform(m2_range[0],m2_range[1])
		elif m2_range is not None:
			m2 = float(m2_range)
		if isinstance(q_range, tuple):
			q = np.random.uniform(q_range[0],q_range[1])
		else:
			q = float(q_range)
		if isinstance(s1_range, tuple):
			spin1z = np.random.uniform(s1_range[0],s1_range[1])
		else:
			spin1z = float(s1_range)
		if isinstance(s2_range, tuple):
			spin2z = np.random.uniform(s2_range[0],s2_range[1])
		else:
			spin2z = float(s2_range)

		if m2_range is None:
			m2 = 20. / (1+q)
			m1 = q * m2
		else:
			m1 = q* m2

			#computing f_min
		f_min = .9* ((151*(t_coal_freq)**(-3./8.) * (((1+q)**2)/q)**(3./8.))/(m1+m2))
		 #in () there is the right scaling formula for frequency in order to get always the right reduced time
		 #this should be multiplied by a prefactor (~1) for dealing with some small variation due to spins

			#getting the wave
		hptilde, hctilde = lalsim.SimInspiralChooseTDWaveform( #where is its definition and documentation????
			m1*lalsim.lal.MSUN_SI, #m1
			m2*lalsim.lal.MSUN_SI, #m2
			0., 0., spin1z, #spin vector 1
			0., 0., spin2z, #spin vector 2
			d*1e6*lalsim.lal.PC_SI, #distance to source
			inclination, #inclination
			0., #phi ref
			0., #longAscNodes
			0., #eccentricity
			0., #meanPerAno
			t_step, # time incremental step
			f_min, # lowest value of time
			f_min, #some reference value of time (??)
			lal.CreateDict(), #some lal dictionary
			lalsim.GetApproximantFromString('SEOBNRv2_opt') #approx method for the model
		)
		#print(f_min, t_step)#debug
		#print(m1,m2, spin1z,spin2z) #debug

		h = np.array(hptilde.data.data)+1j*np.array(hctilde.data.data) #complex waveform
		if isinstance(m2_range, tuple):
			temp_theta = [m1, m2, spin1z, spin2z]		
		else:
			temp_theta = [m1/m2, spin1z, spin2z]
		temp_amp = np.abs(h)
		temp_ph = np.unwrap(np.angle(h))

			#bringing waves on the chosen grid
		time_full = np.linspace(0.0, hptilde.data.length*t_step, hptilde.data.length) #time grid at which wave is computed
		time_full = (time_full - time_full[np.argmax(temp_amp)])/(m1+m2) #grid is scaled to standard grid
			#setting waves to the chosen std grid
		temp_amp = np.interp(time_grid, time_full, temp_amp)
		temp_ph = np.interp(time_grid, time_full, temp_ph)

			#here you need to decide what is better
		#temp_ph = temp_ph - temp_ph[0] #all phases are shifted by a constant to make every wave start with 0 phase
		id0 = np.where(time_grid == 0)[0]
		temp_ph = temp_ph - temp_ph[id0] #all phases are shifted by a constant to make every wave start with 0 phase at t=0 (i.e. at maximum amplitude)

			#removing spourious gaps (if present)
		(index,) = np.where(temp_amp/np.max(temp_amp) < 1e-4) #there should be a way to choose right threshold...
		if len(index) >0:
			print("Wave killed")
			temp_amp[index] = temp_amp[index[0]-1]
			temp_ph[index] = temp_ph[index[0]-1]

		if filename is None:
			amp_dataset[i,:] = temp_amp  #putting waveform in the dataset to return
			ph_dataset[i,:] =  temp_ph  #phase
			theta_vector[i,:] = temp_theta

		if filename is not None: #saving to file
			to_save = np.concatenate((temp_theta, temp_amp, temp_ph))
			to_save = np.reshape(to_save, (1,len(to_save)))
			np.savetxt(filebuff, to_save)

	if filename is None:
		return theta_vector, amp_dataset.real, ph_dataset.real, time_grid
	else:
		filebuff.close()
		return None

def generate_waveform(m1,m2, s1=0.,s2 = 0.,d=1., iota = 0.,phi_0=0., t_coal = 0.4, t_step = 5e-5, f_min = None, t_min = None, verbose = False):
	"""
generate_waveform
=================
	Wrapper to lalsimulation.SimInspiralChooseTDWaveform() to generate a single waveform. Wave is not preprocessed.
	Input:
		m1,m2,s1,s2,d,iota,phi_0	orbital parameters
		t_coal						approximate time to coalescence in reduced grid (ignored if f_min or t_min is set)
		t_step						EOB integration time to be given to lal
		f_min						starting frequency in Hz (if None, it will be determined by t_coal; ignored if t_min is set)
		t_min						starting time in s (if None, t_coal will be returned)
		verbose						whether to print messages for each wave...
	Output:
		times (D,)	times at which wave is evaluated
		h_p (N,D)	plus polarization of the wave
		h_c (N,D)	cross polarization of the wave
	"""
	q = m1/m2
	mtot = (m1+m2)#*lal.MTSUN_SI
	mc = (m1*m2)**(3./5.)/(m1+m2)**(1./5.)
	mc /= 1.21 #M_c / 1.21 M_sun

	if t_min is not None:
		f_min = .8* ((151*(t_min/mtot)**(-3./8.) * (((1+q)**2)/q)**(3./8.))/mtot)

	if f_min is None:
		f_min = .9* ((151*(t_coal)**(-3./8.) * (((1+q)**2)/q)**(3./8.))/mtot)

	if verbose:
		print("Generating wave @: ",m1,m2,s1,s2,d, iota, phi_0)

	hp, hc = lalsim.SimInspiralChooseTDWaveform( #where is its definition and documentation????
		m1*lalsim.lal.MSUN_SI, #m1
		m2*lalsim.lal.MSUN_SI, #m2
		0., 0., s1, #spin vector 1
		0., 0., s2, #spin vector 2
		d*1e6*lalsim.lal.PC_SI, #distance to source
		iota, #inclination
		phi_0, #phi
		0., #longAscNodes
		0., #eccentricity
		0., #meanPerAno
		t_step, # time incremental step
		f_min, # lowest value of freq
		f_min, #some reference value of freq (??)
		lal.CreateDict(), #some lal dictionary
		lalsim.GetApproximantFromString('SEOBNRv2_opt') #approx method for the model
		)
	h_p, h_c =  hp.data.data, hc.data.data
	amp = np.sqrt(h_p**2+h_c**2)
	#(indices, ) = np.where(amp!=0) #trimming zeros of amplitude
	#h_p = h_p[indices]
	#h_c = h_c[indices]

	times = np.linspace(0.0, h_p.shape[0]*t_step, h_p.shape[0])  #time actually
	t_m =  times[np.argmax(amp)]
	times = times - t_m

	if t_min is not None:
		arg = np.argmin(np.abs(times+t_min))
	else:
		arg=0

	return times[arg:], h_p[arg:], h_c[arg:]
	

def create_dataset_FD(N_data, N_grid = None, filename = None, q_range = (1.,5.), m2_range = 20., s1_range = (-0.8,0.8), s2_range = (-0.8,0.8), log_space = True, f_high = 2000, f_step = 1e-2, f_max = None, f_min =None, lal_approximant = "IMRPhenomPv2"):
	"""
	Create a dataset for training a ML model to fit GW waveforms in frequency domain.
	The dataset consists in 3 parameters theta=(q, spin1z, spin2z) associated to the waveform computed in frequency domain for a grid of N_grid points in the range given by the user.
	More specifically, data are stored in 3 vectors:
		theta_vector	vector holding source parameters q, spin1, spin2
		amp_vector		vector holding amplitudes for each source evaluated at some N_grid equally spaced points
		ph_vector		vector holding phase for each source evaluated at some N_grid equally spaced points
	This routine add N_data data to filename if one is specified (if file is not empty it must contain data with the same N_grid); otherwise the datasets are returned as np vectors. 
	All the waves are evaluated at a constant distance of 1Mpc. Values of q and m2 are drawn randomly in the range given by the user: it holds m1 = q *m2 M_sun.
	The waveforms are computed from f_low = 15 to f_high with a step f_step and then evaluated at some N_grid grid points equally spaced in range [f_min, f_max]
	Dataset can be loaded with load_dataset
	Input:
		N_data			size of dataset
		N_grid			number of points to be sampled in the grid (if None every point generated is saved)
		filename		name of the file to save dataset in (If is None, nothing is saved on a file)
		q_range			tuple with range for random q values. if single value, q is kept fixed at that value
		m2_range		tuple with range for random m2 values. if single value, m2 is kept fixed at that value
		spin_mag_max_1	tuple with range for random spin #1 values. if single value, s1 is kept fixed at that value
		spin_mag_max_2	tuple with range for random spin #1 values. if single value, s2 is kept fixed at that value
		log_space		whether grid should be computed in logspace
		f_high			highest frequency to compute
		f_step			step considered for computation of waveforms
		f_max			maximum frequency returned to the user (if None is the same as f_max)
		f_min			minimum frequency returned to the user (if None is the same as f_low = 15)
		lal_approximant	string for the approximant model to be used (in lal convention)
	Output:
		if filename is given
			None
		if filename is not given
			theta_vector (N_data,3)		vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
			amp_dataset (N_data,N_grid)	dataset with amplitudes
			ph_dataset (N_data,N_grid)	dataset with phases
			frequencies (N_grid,)		vector holding frequencies at which waves are evaluated
	"""
	if f_max is None:
		f_max = f_high
	if f_min is None:
		f_min = 1.
	f_low = f_min
	K = int((f_max-f_min)/f_step) #number of data points to be taken from the returned vector
	if N_grid is None:
		N_grid = K
	full_freq = np.arange(f_low, f_max, f_step) #full frequency vector as returned by lal
	d=1.
	LALpars = lal.CreateDict()
	approx = lalsim.SimInspiralGetApproximantFromString(lal_approximant)

		#allocating storage for temp vectors to save a single WF
	temp_amp = np.zeros((N_grid,))
	temp_ph = np.zeros((N_grid,))
	temp_theta = np.zeros((3,))

		#setting frequencies to be returned to user
	if log_space:
		frequencies = np.logspace(np.log10(f_min), np.log10(f_max), N_grid)
	else:
		frequencies = np.linspace(f_min, f_max, N_grid)
		#freq_to_choose = np.arange(0, K, K/N_grid).astype(int) #choosing proper indices s.t. dataset holds N_grid points
		#frequencies = full_freq[freq_to_choose] #setting only frequencies to be chosen

	if filename is not None: #doing header if file is empty - nothing otherwise
		if not os.path.isfile(filename): #file doesn't exist: must be created with proper header
			filebuff = open(filename,'w')
			print("New file ", filename, " created")
			freq_header = np.concatenate((np.zeros((3,)), frequencies, frequencies) )
			freq_header = np.reshape(freq_header, (1,len(freq_header)))
			np.savetxt(filebuff, freq_header, header = "row: theta 3 | amp "+str(frequencies.shape[0])+"| ph "+str(frequencies.shape[0])+"\nN_grid = "+str(N_grid)+" | f_step ="+str(f_step)+" | q_range = "+str(q_range)+" | s1_range = "+str(s1_range)+" | s2_range = "+str(s2_range), newline = '\n')
		else:
			filebuff = open(filename,'a')

	if filename is None:
		amp_dataset = np.zeros((N_data,N_grid)) #allocating storage for returning data
		ph_dataset = np.zeros((N_data,N_grid))
		theta_vector = np.zeros((N_data,3))

	for i in range(N_data): #loop on data to be created
		if i%100 == 0 and i != 0:
			print("Generated WF ", i)

			#setting value for data
		if isinstance(m2_range, tuple):
			m2 = np.random.uniform(m2_range[0],m2_range[1])
		else:
			m2 = m2_range
		if isinstance(q_range, tuple):
			m1 = np.random.uniform(q_range[0],q_range[1]) * m2
		else:
			m1 = q_range * m2
		if isinstance(s1_range, tuple):
			spin1z = np.random.uniform(s1_range[0],s1_range[1])
		else:
			spin1z = s1_range
		if isinstance(s2_range, tuple):
			spin2z = np.random.uniform(s2_range[0],s2_range[1])
		else:
			spin2z = s2_range

			#debug!!!
		if m1/m2 >4.6 and (np.abs(spin1z) >0.8 or np.abs(spin2z) > 0.8):
			continue


			#getting the wave
		hptilde, hctilde = lalsim.SimInspiralChooseFDWaveform( #where is its definition and documentation????
			m1*lalsim.lal.MSUN_SI, #m1
			m2*lalsim.lal.MSUN_SI, #m2
			0., 0., spin1z, #spin vector 1
			0., 0., spin2z, #spin vector 2
			d*1e6*lalsim.lal.PC_SI, #distance to source
			0., #inclination
			0., #phi ref
			0., #longAscNodes (for precession)
			0., #eccentricity
			0., #meanPerAno (for precession)
			f_step, # frequency incremental step
			f_low, # lowest value of frequency
			f_high, # highest value of frequency
			f_low, #some reference value of frequency (??)
			LALpars, #some lal dictionary
			approx #approx method for the model
			)
		h = np.array(hptilde.data.data)+1j*np.array(hctilde.data.data) #complex waveform
		temp_theta = [m1/m2, spin1z, spin2z]
		temp_amp = (np.abs(h)[int(f_min/f_step):int(f_max/f_step)].real)
		temp_ph = (np.unwrap(np.angle(h))[int(f_min/f_step):int(f_max/f_step)].real)

			#bringing waves on the chosen grid
		temp_amp = np.interp(frequencies, full_freq, temp_amp)
		temp_ph = np.interp(frequencies, full_freq, temp_ph)
		#temp_ph = temp_ph[freq_to_choose]; temp_amp = temp_amp[freq_to_choose] #old version of code

		temp_ph = temp_ph - temp_ph[0] #all frequencies are shifted by a constant to make the wave start at zero phase!!!! IMPORTANT

			#removing spourious gaps (if present)
		(index,) = np.where(temp_amp/temp_amp[0] < 5e-3) #there should be a way to choose right threshold...
		if len(index) >0:
			temp_ph[index] = temp_ph[index[0]-1]

		if filename is None:
			amp_dataset[i,:] = temp_amp  #putting waveform in the dataset to return
			ph_dataset[i,:] =  temp_ph  #phase
			theta_vector[i,:] = temp_theta

		if filename is not None: #saving to file
			to_save = np.concatenate((temp_theta,temp_amp, temp_ph))
			to_save = np.reshape(to_save, (1,len(to_save)))
			np.savetxt(filebuff, to_save)

	if filename is None:
		return theta_vector, amp_dataset.real, ph_dataset.real, frequencies
	else:
		filebuff.close()
		return None

def save_dataset(filename, theta_vector, dataset1, dataset, x_grid):
	"""
	Save a dataset in a way that it's readable by load_dataset.
	Input:
		filename	name of the file to save dataset to
		theta_vector (N_data,3)		vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
		amp_dataset (N_data,N_grid)	dataset with amplitudes
		ph_dataset (N_data,N_grid)	dataset with phases
		x_grid (N_grid,)			vector holding x_grid at which waves are evaluated
	"""
	to_save = np.concatenate((theta_vector, amp_dataset, ph_dataset), axis = 1)
	temp_x_grid = np.zeros((1,to_save.shape[1]))
	K = int((to_save.shape[1]-3)/2)
	temp_x_grid[0,3:3+k] = x_grid
	to_save = np.concatenate((temp_x_grid,to_save), axis = 0)
	q_max = np.max(theta_vector[:,0])
	spin_mag_max = np.max(np.abs(theta_vector[:,1:2]))
	x_step = x_grid[1]-x_grid[0]
	np.savetxt(filename, to_save, header = "row: theta 3 | amp "+str(amp_dataset.shape[1])+"| ph "+str(ph_dataset.shape[1])+"\nN_grid = "+str(x_grid.shape[0])+" | f_step ="+str(x_step)+" | q_max = "+str(q_max)+" | spin_mag_max = "+str(spin_mag_max), newline = '\n')
	return


def load_dataset(filename, N_data=None, N_grid = None, shuffle = False):
	"""
	Load a GW dataset from file. The file should be suitable for np arrays and have the following structure:
		theta 3 | amplitudes K | phases K
	The first row hold the frequncy vector.
	It can shuffle the data if required.
	Input:
		filename	input filename
		N_data		number of data to extract (only if data in file are more than N_data) (if None N_data = N)
		N_grid		number of grid points to evaluate the waves in (Only if N_grid < N_grid_dataset)
		shuffle		whether to shuffle data
	Outuput:
		theta_vector (N_data,3)	vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
		amp_dataset (N_data,K)	dataset with amplitudes and wave parameters K = (f_high-30)/(f_step*N_grid)
		ph_dataset (N_data,K)	dataset with phases and wave parameters K = (f_high-30)/(f_step*N_grid)
		x_grid (K,)				vector holding x_grid at which waves are evaluated (can be frequency or time grid)
	"""
	if N_data is not None:
		N_data += 1
	data = np.loadtxt(filename, max_rows = N_data)
	N = data.shape[0]
	K = int((data.shape[1]-3)/2)
	x_grid = data[0,3:3+K] #saving x_grid

	data = data[1:,:] #removing x_grid from full data
	if shuffle: 	#shuffling if required
		np.random.shuffle(data)

	theta_vector = data[:,0:3]
	amp_dataset = data[:,3:3+K]
	ph_dataset = data[:,3+K:]

	if N_grid is not None:
		if ph_dataset.shape[1] < N_grid:
			print("Not enough grid points ("+str(ph_dataset.shape[1])+") for the required N_grid value ("+str(N_grid)+").\nMaximum number of grid point is taken (but less than N_grid)")
			N_grid = ph_dataset.shape[1]
		indices = np.arange(0, ph_dataset.shape[1], int(ph_dataset.shape[1]/N_grid)).astype(int)
		x_grid = x_grid[indices]
		ph_dataset = ph_dataset[:,indices]
		amp_dataset = amp_dataset[:,indices]

	return theta_vector, amp_dataset, ph_dataset, x_grid

def make_set_split(data, labels, train_fraction = .85, scale_factor = None):
	"""
	Given a GW dataset made of data and labels, it makes a split between training and test set. Labels are scaled for scale factor (labels = labels/scale_factor).
	Input:
		data (N,K)		parameters set
		labels (N,L)	label set
		train_fraction	the fraction of data to included in training set
		scale_factor	scale factor for scaling data (if None data are not scaled)
	Output:
		train_data/test_data (N*train_frac/N*(1-train_frac),K)	parameters for training/test set
		train_labels/test_labels (N*train_frac/N*(1-train_frac),K)	labels for training/test set
	"""
	if scale_factor is None:
		scale_factor = 1.
	labels = labels/scale_factor

		#splitting into train and test set
	N = data.shape[0]
	train_data = np.array(data[:int(train_fraction*N),:])
	test_data = np.array(data[int(train_fraction*N):,:])

	train_labels = np.array(labels[:int(train_fraction*N),:])
	test_labels = np.array(labels[int(train_fraction*N):,:])
	return train_data, test_data, train_labels, test_labels

def process_amplitudes(frequencies, q_vector, amp_dataset, get_true_amp):
	"""
	Process an amplitude dataset by scaling it according to:
		A' = A * (f**(7/6) M_c**(-5/6))
	It computes A if get_true_amp is true; A' otherwise
	Input:
		frequencies (D,)	frequencies at which amplitudes are evaluated
		q_vector (N,)		vector holding mass ratio q for every GW in the dataset
		amp_dataset (N,D)	N samples of amplitudes each evaluated at the D point in frequencies
		get_true_amp		whether to compute A (computes A if True, A' if false)
	Output:
		new_amp (N,D)		transformed amplitude dataset
	"""
	N = amp_dataset.shape[0]
	D = amp_dataset.shape[1]
	M_c = np.reshape(np.divide(np.power(q_vector,3./5.), np.power(1.+q_vector,1./5.)), (N,1)) #chirp mass
	new_frequencies = np.array(frequencies)
	new_frequencies = np.repeat(np.reshape(new_frequencies, (1,D) ), N, axis=0)
	factor = np.multiply(np.power(new_frequencies, 7./6.), np.power(M_c, -5./6.)) * 1e19
	if get_true_amp is False:
		return np.multiply(amp_dataset, factor)
	if get_true_amp is True:
		return np.divide(amp_dataset, factor)

def process_phases(frequencies, q_vector, ph_dataset, get_true_ph):
	"""
	Process a phase dataset by scaling it according to:
		ph' = ph + offset
	It computes ph' if get_true_ph is true; ph' otherwise
	Input:
		frequencies (D,)	frequencies at which amplitudes are evaluated
		q_vector (N,)		vector holding mass ratio q for every GW in the dataset
		ph_dataset (N,D)	N samples of amplitudes each evaluated at the D point in frequencies
		get_true_ph			whether to compute A (computes A if True, A' if false)
	Output:
		new_ph (N,D)	transformed amplitude dataset
	"""
	N = ph_dataset.shape[0]
	D = ph_dataset.shape[1]
	new_frequencies = np.array(frequencies)

	M_c = np.reshape(np.divide(np.power(q_vector,3./5.), np.power(1.+q_vector,1./5.)), (N,1)) #chirp mass
	m = 1.+q_vector
	nu = np.divide(q_vector, np.power(m,2.))
	prefactor = 6.6e-11 * 2e30 /(3e8)**3 #G*M_sun/c**3
	M = prefactor*m
		#PN coefficients... (p.298 Maggiore)
		#t_i (N,)
	t_0 = np.divide(np.power(M*np.pi,-5./3.), nu) * (5./(256*np.pi))
	t_1 = np.multiply(np.power(np.multiply(M*np.pi,nu),-1.), 743./336 + 11/4. * nu) * (5./(192*np.pi))
	t_1_5 = np.divide(np.power(M*np.pi,-2./3.), nu) / 8.
	t_2 =  np.multiply(np.divide(np.power(M*np.pi,-1./3.), nu), 3058673./1016064. + 5429/1008 * nu+ 617/144 * np.square(nu))  * (5./(128*np.pi))
	
	for i in range(N):
		offset = (3/5. * t_0[i] * np.power(new_frequencies,-5/3.) + t_1[i] * np.power(new_frequencies,-1.) - 1.5 *t_1_5[i] * np.power(new_frequencies,-2/3.) + 3 * t_2[i] * np.power(new_frequencies,-1./3.) )* 2*np.pi
		#ph_dataset[i,:] = ph_dataset[i,:]+offset-np.max(ph_dataset[i,:])+100

	offset = (3/5. * np.outer(t_0, np.power(new_frequencies,-5/3.)) + np.outer(t_1, np.power(new_frequencies,-1.)) - 1.5 *np.outer(t_1_5, np.power(new_frequencies,-2/3.)) + 3 *np.outer(t_2, np.power(new_frequencies,-1./3.)) )* 2*np.pi
	offset = offset + 100
	#print(np.max(ph_dataset, axis =1)[0:5])

	if get_true_ph is False:
		return ph_dataset +offset
	if get_true_ph is True:
		return ph_dataset - offset



#################################OLD CODE#######################################
def create_dataset_red_f(N_data, N_grid = 3000, filename = None, q_max = 18, spin_mag_max = 0.9, w_min = .1, w_max = 1., lal_approximant = "IMRPhenomD"):
	"""
	Create a dataset for training a ML model to fit GW waveforms.
	The dataset consists in 3 parameters theta=(q, spin1z, spin2z) associated to the waveform computed in frequency domain for a grid of N_grid points in the range given by the user.
	More specifically, data are stored in 3 vectors:
		theta_vector	vector holding source parameters q, spin1, spin2
		amp_vector		vector holding amplitudes for each source evaluated at some N_grid equally spaced points
		ph_vector		vector holding phase for each source evaluated at some N_grid equally spaced points
	This routine add N_data data to filename if one is specified (if file is not empty it must contain data with the same N_grid); otherwise the datasets are returned as np vectors. 
	All the waves are evaluated at a constant distance of 1Mpc and with m2 = 10 M_sun and m1 = q M_sun.
	The waveforms are computed on a fixed grid of reduced (dimensionless) frequency
		w = scale * f =(m1+m2)*2.0*np.pi * f
	at some N_grid points equally spaced in range [w_min, w_max]. Phases are divided by the factor scale**(-5/3).
	Those choices of preprocessing make the dataset more omogenous since every signal has the similar lenght and magnitude O(1).
	Dataset can be loaded with load_dataset
	Input:
		N_data			size of dataset
		N_grid			number of points to be sampled in the grid (if None every point generated is saved)
		filename		name of the file to save dataset in (If is None, nothing is saved on a file)
		q_max			maximum mass ratio q to be considered (>1)
		spin_mag_max	maximum spin magnitude to be considered (>0) (dangerous: for high q actual spin_mag_max is lower)
		f_high			highest frequency to do compute
		f_step			step considered for computation of waveforms
		f_max			maximum frequency returned to the user (if None is the same as f_max)
		f_min			minimum frequency returned to the user (if None is the same as f_low = 15)
		lal_approximant	string for the approximant model to be used (in lal convention)
	Output:
		if filename is given
			None
		if filename is not given
			theta_vector (N_data,3)		vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
			amp_dataset (N_data,N_grid)	dataset with amplitudes
			ph_dataset (N_data,N_grid)	dataset with phases
			frequencies (N_grid,)		vector holding frequencies at which waves are evaluated
	"""
	wstd = np.linspace(w_min, w_max, N_grid) #reduced grid in which every point is evaluated (returned)
	dw= ((w_max-w_min)/N_grid)
	spin_mag_max = np.abs(spin_mag_max)
	m2 = 10.
	d=1.
	LALpars = lal.CreateDict()
	approx = lalsim.SimInspiralGetApproximantFromString(lal_approximant)

		#allocating storage for temp vectors to save a single WF
	temp_amp = np.zeros((N_grid,))
	temp_ph = np.zeros((N_grid,))
	temp_theta = np.zeros((3,))

	if filename is not None: #doing header if file is empty - nothing otherwise
		if not os.path.isfile(filename): #file doesn't exist: must be created with proper header
			filebuff = open(filename,'w')
			print("New file ", filename, " created")
			freq_header = np.concatenate((np.zeros((3,)), wstd, wstd) )
			freq_header = np.reshape(freq_header, (1,len(freq_header)))
			np.savetxt(filebuff, freq_header, header = "row: theta 3 | amp "+str(N_grid)+"| ph "+str(N_grid)+"\nN_grid = "+str(N_grid)+" | f_red_min ="+str(w_min)+" | f_red_max ="+str(w_max)+ " | q_max = "+str(q_max)+" | spin_mag_max = "+str(spin_mag_max), newline = '\n')
		else:
			filebuff = open(filename,'a')

	if filename is None:
		amp_dataset = np.zeros((N_data,N_grid)) #allocating storage for returning data
		ph_dataset = np.zeros((N_data,N_grid))
		theta_vector = np.zeros((N_data,3))

	for i in range(N_data): #loop on data to be created
		if i%100 == 0 and i != 0:
			print("Generated WF ", i)
		m1 = 	np.random.uniform(1,q_max) * m2
		spin1z = np.random.uniform(-spin_mag_max,spin_mag_max)
		spin2z = np.random.uniform(-spin_mag_max,spin_mag_max)
		scale = (m1+m2)*lal.MTSUN_SI*2.0*np.pi #frequency scale factor [1/Hz]: f_red = w = f*scale

			#params for wave generation (given total mass)
		f_high = w_max/scale
		f_low = 20.#w_min/scale
		w_min = scale*20 #debug
		wstd = np.linspace(w_min, w_max, N_grid) #debug
		#print(m1+m2,scale,f_low, f_high)
		f_step = dw/scale

			#getting the wave
		hptilde, hctilde = lalsim.SimInspiralChooseFDWaveform( #where is its definition and documentation????
			m1*lalsim.lal.MSUN_SI, #m1
			m2*lalsim.lal.MSUN_SI, #m2
			0, 0, spin1z, #spin vector 1
			0, 0, spin2z, #spin vector 2
			d*1e6*lalsim.lal.PC_SI, #distance to source
			0, #inclination
			0, #something I don't know
			0, #longAscNodes
			0, #eccentricity
			0, #meanPerAno
			f_step, # frequency incremental step
			f_low, # lowest value of frequency
			f_high, # highest value of frequency
			f_low, #some reference value of frequency (??)
			LALpars, #some lal dictionary
			approx #approx method for the model
			)

		hpt = hptilde.data.data
		hct = hctilde.data.data
		ftmp = w_min/scale + dw*np.arange(hpt.shape[0])/scale #grid in which WF are generated
		fstd = wstd /scale #frequencies s.t. the WF in reduced frequencies is evaluated at wstd grid

		hp = np.interp(fstd,ftmp,hpt) #bringing waves to fstd grid
		hc = np.interp(fstd,ftmp,hct)

		temp_theta = [m1/m2, spin1z, spin2z]
		temp_amp = np.abs(hp+1j*hc).real
		temp_ph = np.unwrap(np.angle(hp+1j*hc)).real/scale**(-5./3.) #phase is scaled by scale**(-5/3)

		if filename is None:
			amp_dataset[i,:] = temp_amp  #putting waveform in the dataset to return
			ph_dataset[i,:] =  temp_ph  #phase
			theta_vector[i,:] = temp_theta

		if filename is not None: #saving to file
			to_save = np.concatenate((temp_theta,temp_amp, temp_ph))
			to_save = np.reshape(to_save, (1,len(to_save)))
			#print(to_save.shape, temp_amp.shape)
			np.savetxt(filebuff, to_save)

	if filename is None:
		return theta_vector, amp_dataset.real, ph_dataset.real, wstd
	else:
		filebuff.close()
		return None	

def transform_dataset(theta_vector, amp_dataset, ph_dataset, w_vector, f_vector, set_w_grid):
	"""
	Given a dataset (amplitudes vector and phase vector) in some grid, it returns the same WFs evaluated on another grid. The two grids must be of different kinds (there must be a w_grid and a f_grid). It performs as well the suitable scaling in the phase magnitude.
	Input:
		theta_vector (N_data,3)	vector holding ordered set of parameters used to generate amp_dataset and ph_dataset
		amp_dataset (N_data,K)	dataset with amplitudes and wave parameters K = (f_high-30)/(f_step*N_grid)
		ph_dataset (N_data,K)	dataset with phases and wave parameters K = (f_high-30)/(f_step*N_grid)
		w_vector (K,)			vector holding frequencies at which waves are evaluated in w_grid
		f_vector (K',)			vector holding frequencies at which waves are evaluated in f_grid
		set_w_grid (bool)		if true data are brought from f_grid to w_grid; viceversa if false
	Output:
		new_amp_dataset (N_data,K')	new dataset with amplitudes
		new_ph_dataset (N_data,K')	new dataset with phases
	"""
	if set_w_grid:
		N_grid_new = w_vector.shape[0]
	else:
		N_grid_new = f_vector.shape[0]

	new_amp_dataset = np.zeros((amp_dataset.shape[0], N_grid_new))
	new_ph_dataset = np.zeros((ph_dataset.shape[0], N_grid_new))

	for i in range(amp_dataset.shape[0]):
		scale = (theta_vector[i,0]+1)*10*lal.MTSUN_SI*2.0*np.pi #frequency scale factor [1/Hz]
		wtmp = w_vector / scale
		if set_w_grid:
			new_amp_dataset[i,:] = np.interp(wtmp, f_vector, amp_dataset[i,:])
			new_ph_dataset[i,:] = np.interp(wtmp, f_vector, ph_dataset[i,:]) /scale**(-5./3.)
		else:
			new_amp_dataset[i,:] = np.interp(f_vector, wtmp, amp_dataset[i,:])
			#print(f_vector, wtmp)
			new_ph_dataset[i,:] = np.interp(f_vector, wtmp, ph_dataset[i,:])*scale**(-5./3.)

	return new_amp_dataset, new_ph_dataset



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='mlgw',
    version='1.0.post3',
    author='Stefano Schmidt',
    author_email='stefanoschmidt1995@gmail.com',
    packages=['mlgw'],
    package_dir = {'mlgw':'./mlgw'},
    url="https://github.com/stefanoschmidt1995/MLGW/",
    license='CC by 4.0',
    description='Machine learning model for the gravitational waves generated by black-hole binaries',
    long_description=readme(),
    include_package_data = True,
    package_data={'mlgw': ['TD_model/*']},
    install_requires=[
        "numpy >= 1.16.4",
		"scipy >= 1.3.1",
		"lalsuite >= 6.62"
    ],
)

